from contextlib import contextmanager
from io import BytesIO
import logging
import os
from pathlib import Path
import tempfile
from typing import Optional, Tuple, Union

from .models import Artist

from requests import Session

TIDAL_API_URL: str = "https://api.tidal.com/v1"
IMAGE_URL: str = "https://resources.tidal.com/images/%s.jpg"

logger = logging.getLogger(__name__)


def download_cover_image(
    session: Session,
    cover_uuid: str,
    output_dir: Path,
    file_name: str = "cover.jpg",
    dimension: Union[int, Tuple[int]] = 1280,
) -> Optional[Path]:
    """Given a UUID that corresponds to a (JPEG) image on Tidal's servers,
    download the image file and write it as 'cover.jpeg' or 'cover.png'
    in the directory `path_to_output_dir`. Returns path to downloaded file"""
    cover_url_part: str = cover_uuid.replace("-", "/")
    if isinstance(dimension, int):
        _url: str = IMAGE_URL % f"{cover_url_part}/{dimension}x{dimension}"
    elif isinstance(dimension, tuple):
        _url: str = IMAGE_URL % f"{cover_url_part}/{dimension[0]}x{dimension[1]}"

    with session.get(url=_url, headers={"Accept": "image/jpeg"}) as r:
        if not r.ok:
            logger.warning(
                "Could not retrieve data from Tidal resources/images URL "
                f"due to error code: {r.status_code}"
            )
            logger.debug(r.reason)
            return
        else:
            bytes_to_write = BytesIO(r.content)

    if bytes_to_write is not None:
        output_file: Path = output_dir / file_name
        bytes_to_write.seek(0)
        output_file.write_bytes(bytes_to_write.read())
        bytes_to_write.close()
        return output_file


def download_artist_image(
    session: Session, artist: Artist, output_dir: Path, dimension: int = 320
) -> Optional[Path]:
    """Given a UUID that corresponds to a (JPEG) image on Tidal's servers,
    download the image file and write it as '{artist name}.jpeg'
    in the directory `output_dir`. Returns path to downloaded file"""
    _url: str = artist.picture_url(dimension)
    if _url is None:
        logger.info(
            f"Cannot download image for artist '{artist}', "
            "as Tidal supplied no URL for this artist's image."
        )
        return

    with session.get(url=_url, headers={"Accept": "image/jpeg"}) as r:
        if not r.ok:
            logger.warning(
                "Could not retrieve data from Tidal resources/images URL "
                f"for artist {artist} due to error code: {r.status_code}"
            )
            logger.debug(r.reason)
            return
        else:
            bytes_to_write = BytesIO(r.content)

    file_name: str = f"{artist.name.replace('..', '')}.jpg"
    if bytes_to_write is not None:
        output_file: Path = output_dir / file_name
        bytes_to_write.seek(0)
        output_file.write_bytes(bytes_to_write.read())
        bytes_to_write.close()
        logger.info(
            f"Wrote artist image JPEG for {artist} to "
            f"'{str(output_file.absolute())}'"
        )
        return output_file


@contextmanager
def temporary_file(suffix: str = ".mka"):
    """This context-managed function is a stand-in for
    tempfile.NamedTemporaryFile as that stdlib object experiences
    errors on Windows."""
    file_name: str = os.path.join(
        tempfile.gettempdir(), f"{os.urandom(24).hex()}{suffix}"
    )
    if not os.path.exists(file_name):
        open(file=file_name, mode="x").close()

    tf = open(file=file_name, mode="wb")
    try:
        yield tf
    finally:
        tf.close()
        os.unlink(tf.name)



import base64
from enum import Enum
import json
import logging
from pathlib import Path
import sys
from typing import Dict, Optional, Set, Tuple

from .models import BearerAuth, SessionsEndpointResponseJSON
from .oauth import (
    TOKEN_DIR_PATH,
    BearerToken,
    TidalOauth,
    TokenException,
)
from .utils import TIDAL_API_URL

import requests
import typer

COMMON_HEADERS: Dict[str, str] = {"Accept-Encoding": "gzip, deflate, br"}

logger = logging.getLogger(__name__)


class AudioFormat(str, Enum):
    sony_360_reality_audio = "360"
    dolby_atmos = "Atmos"
    hi_res = "HiRes"
    mqa = "MQA"
    lossless = "Lossless"
    high = "High"
    low = "Low"


class LogLevel(str, Enum):
    debug = "DEBUG"  # 10
    info = "INFO"  # 20
    warning = "WARNING"  # 30
    error = "ERROR"  # 40
    critical = "CRITICAL"  # 50


def load_token_from_disk(
    token_path: Path = TOKEN_DIR_PATH / "android-tidal.token",
) -> Optional[str]:
    """Attempt to read `token_path` from disk and decoded its contents
    as JSON"""
    if not token_path.exists():
        logger.warning(f"FileNotFoundError: {str(token_path.absolute())}")
        return
    token_file_contents: str = token_path.read_bytes()
    decoded_token_file_contents: str = base64.b64decode(token_file_contents).decode(
        "utf-8"
    )

    try:
        bearer_token_json: dict = json.loads(decoded_token_file_contents)
    except json.decoder.JSONDecodeError:
        logger.warning(f"File '{token_path.absolute()}' cannot be parsed as JSON")
        return
    else:
        return bearer_token_json.get("access_token")


def validate_token(
    token: str, headers: Dict[str, str] = COMMON_HEADERS
) -> Optional[requests.Session]:
    """Send a GET request to the /sessions endpoint of Tidal's API.
    If `token` is valid, use the SessionsEndpointResponseJSON object
    that was returned from the API to create a requests.Session object with
    some additional attributes. Otherwise, return None"""
    auth_headers: Dict[str, str] = {**headers, "Authorization": f"Bearer {token}"}
    sess: Optional[requests.Session] = None

    with requests.get(url=f"{TIDAL_API_URL}/sessions", headers=auth_headers) as r:
        try:
            r.raise_for_status()
        except requests.HTTPError as h:
            if r.status_code == 401:
                logger.error("Token is not authorized")
                return sess
            else:
                logger.exception(h)
                return sess

        serj = SessionsEndpointResponseJSON.from_dict(r.json())
        logger.debug("Adding data from API reponse to session object:")
        logger.debug(serj)

    sess: requests.Session = requests.Session()
    sess.headers: Dict[str, str] = headers
    sess.auth: BearerAuth = BearerAuth(token=token)
    sess.user_id: str = serj.user_id
    sess.session_id: str = serj.session_id
    sess.client_id: str = serj.client.id
    sess.client_name: str = serj.client.name
    if serj.country_code == "US":
        sess.params["countryCode"] = "US"
        sess.params["locale"] = "en_US"
        sess.headers["Accept-Language"] = "en-US"
    elif (len(serj.country_code) == 2) and (serj.country_code.isupper()):
        sess.params["countryCode"] = serj.country_code
    return sess


def login_fire_tv(
    token_path: Path = TOKEN_DIR_PATH / "fire_tv-tidal.token",
) -> Optional[requests.Session]:
    """Load `token_path` from disk, initializing a BearerToken from its
    contents. If successful, return a requests.Session object with
    extra attributes set
    """
    bearer_token = BearerToken.load(p=token_path)
    if bearer_token is not None:
        bearer_token.save(p=token_path)
    else:
        to = TidalOauth()
        bearer_token = to.authorization_code_flow()
        logger.info("Successfully loaded token from disk.")
        bearer_token.save(p=token_path)

    # check if access needs refreshed
    if bearer_token.is_expired:
        logger.warning("TIDAL access token needs refreshing: Attempting now.")
        try:
            bearer_token.refresh()
        except TokenException as te:
            sys.exit(te.args[0])
        else:
            logger.info("Successfully refreshed TIDAL access token")

    s: Optional[requests.Session] = validate_token(bearer_token.access_token)
    if s is None:
        logger.critical("Access token is not valid: exiting now.")
    else:
        s.params["deviceType"] = "TV"
        s.headers["User-Agent"] = "TIDAL_ANDROID/2.38.0"
        bearer_token.save()
    return s


def login_android(
    token_path: Path = TOKEN_DIR_PATH / "android-tidal.token",
) -> Optional[requests.Session]:
    logger.info(f"Loading TIDAL access token from '{str(token_path.absolute())}'")
    _token: Optional[str] = load_token_from_disk(token_path=token_path)
    device_type: Optional[str] = None

    if _token is None:
        logger.warning("Could not load access token from disk")
        _token: str = typer.prompt(
            "Enter TIDAL access token from an Android (the part after 'Bearer ')"
        )
        dt_input: str = typer.prompt(
            "Is this from device type: phone, tablet, or other? "
        )
        if dt_input.lower() == "phone":
            device_type = "PHONE"
        elif dt_input.lower() == "tablet":
            device_type = "TABLET"

    s: Optional[requests.Session] = validate_token(_token)
    if s is None:
        logger.critical("Access token is not valid: exiting now.")
        if token_path.exists():
            token_path.unlink()
    else:
        logger.info(f"Access token is valid: saving to {str(token_path.absolute())}")
        if device_type is not None:
            s.params["deviceType"] = device_type

        s.headers["User-Agent"] = "TIDAL_ANDROID/1136 okhttp 4.3.0"
        to_write: dict = {
            "access_token": s.auth.token,
            "session_id": s.session_id,
            "client_id": s.client_id,
            "client_name": s.client_name,
            "country_code": s.params["countryCode"],
        }
        logger.debug(f"Writing this bearer token to '{str(token_path.absolute())}'")
        token_path.write_bytes(base64.b64encode(bytes(json.dumps(to_write), "UTF-8")))
    return s


def login_windows(
    token_path: Path = TOKEN_DIR_PATH / "windows-tidal.token",
) -> Optional[requests.Session]:
    _token: Optional[str] = load_token_from_disk(token_path=token_path)
    if _token is None:
        _token: str = typer.prompt(
            "Enter Tidal API access token (the part after 'Bearer ')"
        )

    s: Optional[requests.Session] = validate_token(_token)
    if s is None:
        logger.critical("Access token is not valid: exiting now.")
    else:
        logger.debug(f"Writing this access token to '{str(token_path.absolute())}'")
        s.headers["User-Agent"] = "TIDAL_NATIVE_PLAYER/WIN/3.1.2.195"
        s.params["deviceType"] = "DESKTOP"
        to_write: dict = {
            "access_token": s.auth.token,
            "session_id": s.session_id,
            "client_id": s.client_id,
            "client_name": s.client_name,
            "country_code": s.params["countryCode"],
        }
        token_path.write_bytes(base64.b64encode(bytes(json.dumps(to_write), "UTF-8")))
    return s


def login_mac_os(
    token_path: Path = TOKEN_DIR_PATH / "mac_os-tidal.token",
) -> Optional[requests.Session]:
    raise NotImplementedError


def login(
    audio_format: AudioFormat,
) -> Tuple[Optional[requests.Session], Optional[AudioFormat]]:
    """Given a selected audio_format, either log in "automatically"
    via the Fire TV OAuth 2.0 flow, or ask for an Android-/Windows-/MacOS-
    gleaned API token; the latter to be able to access HiRes fLaC audio.
    Returns a tuple of a requests.Session object, if no error, and the
    AudioFormat instance passed in; or (None, "") in the event of error.
    """
    android_formats: Set[AudioFormat] = {
        AudioFormat.sony_360_reality_audio,
        AudioFormat.hi_res,
    }
    fire_tv_formats: Set[AudioFormat] = {
        AudioFormat.dolby_atmos,
        AudioFormat.mqa,
        AudioFormat.lossless,
        AudioFormat.high,
        AudioFormat.low,
    }
    if audio_format in fire_tv_formats:
        return (login_fire_tv(), audio_format)
    elif audio_format in android_formats:
        options: set = {"android", "a", "windows", "w"}
        _input: str = ""
        while _input not in options:
            _input = typer.prompt(
                "For which of Android [a] or Windows [w] would you like to provide an API token?"
            ).lower()
        else:
            if _input in {"android", "a"}:
                return (login_android(), audio_format)
            elif _input in {"windows", "w"}:
                return (login_windows(), audio_format)
    else:
        logger.critical(
            "Please provide one of the following: "
            f"{', '.join(e.value for e in AudioFormat)}"
        )
        return (None, "")



from contextlib import closing
import logging
from pathlib import Path
from typing import Optional, Union

from .login import login, AudioFormat, LogLevel
from .album import Album
from .artist import Artist
from .mix import Mix
from .playlist import Playlist
from .track import Track
from .video import Video
from .models import (
    match_tidal_url,
    TidalAlbum,
    TidalArtist,
    TidalMix,
    TidalPlaylist,
    TidalTrack,
    TidalVideo,
)

from platformdirs import user_music_path
import typer
from typing_extensions import Annotated

app = typer.Typer()


@app.command()
def main(
    tidal_url: Annotated[
        str,
        typer.Argument(
            help="The Tidal album or artist or mix or playlist or track or video to download"
        ),
    ],
    audio_format: Annotated[
        AudioFormat, typer.Option(case_sensitive=False)
    ] = AudioFormat.lossless.value,
    output_directory: Annotated[
        Path,
        typer.Argument(
            help="The parent directory under which directory(ies) of files will be written"
        ),
    ] = user_music_path(),
    loglevel: Annotated[
        LogLevel, typer.Option(case_sensitive=False)
    ] = LogLevel.info.value,
    include_eps_singles: Annotated[
        bool,
        typer.Option(
            "--include-eps-singles",
            help="No-op unless passing TIDAL artist. Whether to include artist's EPs and singles with albums",
        ),
    ] = False,
):
    logging.basicConfig(
        format="%(asctime)s,%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s",
        datefmt="%Y-%m-%d:%H:%M:%S",
        level=logging.getLevelName(loglevel.value),
    )
    logger = logging.getLogger(__name__)

    tidal_resource: Optional[
        Union[TidalAlbum, TidalMix, TidalPlaylist, TidalTrack, TidalVideo]
    ] = match_tidal_url(tidal_url)

    if tidal_resource is None:
        logger.critical(
            f"Cannot parse '{tidal_url}' as a TIDAL album, artist, mix, playlist, track, or video URL"
        )
        raise typer.Exit(code=1)

    s, audio_format = login(audio_format=audio_format)
    if s is None:
        raise typer.Exit(code=1)

    with closing(s) as session:
        if isinstance(tidal_resource, TidalTrack):
            track = Track(track_id=tidal_resource.tidal_id)
            track.get(
                session=session, audio_format=audio_format, out_dir=output_directory
            )

            if loglevel == LogLevel.debug:
                track.dump()
            raise typer.Exit(code=0)
        elif isinstance(tidal_resource, TidalAlbum):
            album = Album(album_id=tidal_resource.tidal_id)
            album.get(
                session=session, audio_format=audio_format, out_dir=output_directory
            )

            if loglevel == LogLevel.debug:
                album.dump()
            raise typer.Exit(code=0)
        elif isinstance(tidal_resource, TidalArtist):
            artist = Artist(artist_id=tidal_resource.tidal_id)
            artist.get(
                session=session,
                audio_format=audio_format,
                out_dir=output_directory,
                include_eps_singles=include_eps_singles,
            )
            raise typer.Exit(code=0)
        elif isinstance(tidal_resource, TidalVideo):
            video = Video(video_id=tidal_resource.tidal_id)
            video.get(session=session, out_dir=output_directory)

            if loglevel == LogLevel.debug:
                video.dump()
            raise typer.Exit(code=0)
        elif isinstance(tidal_resource, TidalPlaylist):
            playlist = Playlist(playlist_id=tidal_resource.tidal_id)
            playlist.get(
                session=session, audio_format=audio_format, out_dir=output_directory
            )

            if loglevel == LogLevel.debug:
                playlist.dump()
            raise typer.Exit(code=0)
        elif isinstance(tidal_resource, TidalMix):
            mix = Mix(mix_id=tidal_resource.tidal_id)
            mix.get(
                session=session, audio_format=audio_format, out_dir=output_directory
            )

            if loglevel == LogLevel.debug:
                mix.dump()
            raise typer.Exit(code=0)
        else:
            raise NotImplementedError


if __name__ == "__main__":
    app()



from enum import Enum
from typing import Dict


class AudioFormat(str, Enum):
    sony_360_reality_audio = "360"
    dolby_atmos = "Atmos"
    hi_res = "HiRes"
    mqa = "MQA"
    lossless = "Lossless"
    high = "High"
    low = "Low"


class VideoFormat(str, Enum):
    high = "HIGH"
    medium = "MEDIUM"
    low = "LOW"
    audio_only = "AUDIO_ONLY"


af_aq: Dict[AudioFormat, str] = {
    AudioFormat.sony_360_reality_audio: "LOW",
    AudioFormat.dolby_atmos: "LOW",
    AudioFormat.hi_res: "HI_RES",
    AudioFormat.mqa: "HI_RES",
    AudioFormat.lossless: "LOSSLESS",
    AudioFormat.high: "HIGH",
    AudioFormat.low: "LOW",
}

TAG_MAPPING: Dict[str, Dict[str, str]] = {
    "album": {"flac": "ALBUM", "m4a": "\xa9alb"},
    "album_artist": {"flac": "ALBUMARTIST", "m4a": "aART"},
    "artist": {"flac": "ARTIST", "m4a": "\xa9ART"},
    "artists": {"flac": "ARTISTS", "m4a": "----:com.apple.iTunes:ARTISTS"},
    "barcode": {"flac": "BARCODE", "m4a": "----:com.apple.iTunes:BARCODE"},
    "comment": {"flac": "COMMENT", "m4a": "\xa9cmt"},
    "composer": {"flac": "COMPOSER", "m4a": "\xa9wrt"},
    "copyright": {"flac": "COPYRIGHT", "m4a": "cprt"},
    "date": {"flac": "DATE", "m4a": "\xa9day"},
    "director": {"flac": None, "m4a": "\xa9dir"},
    "engineer": {"flac": "ENGINEER", "m4a": "----:com.apple.iTunes:ENGINEER"},
    "isrc": {"flac": "ISRC", "m4a": "----:com.apple.iTunes:ISRC"},
    "lyrics": {"flac": "LYRICS", "m4a": "\xa9lyr"},
    "lyricist": {"flac": "LYRICIST", "m4a": "----:com.apple.iTunes:LYRICIST"},
    "mixer": {"flac": "MIXER", "m4a": "----:com.apple.iTunes:MIXER"},
    "producer": {"flac": "PRODUCER", "m4a": "----:com.apple.iTunes:PRODUCER"},
    "remixer": {"flac": "REMIXER", "m4a": "----:com.apple.iTunes:REMIXER"},
    "album_peak_amplitude": {
        "flac": "REPLAYGAIN_ALBUM_PEAK",
        "m4a": "----:com.apple.iTunes:REPLAYGAIN_ALBUM_PEAK",
    },
    "album_replay_gain": {
        "flac": "REPLAYGAIN_ALBUM_GAIN",
        "m4a": "----:com.apple.iTunes:REPLAYGAIN_ALBUM_GAIN",
    },
    "track_peak_amplitude": {
        "flac": "REPLAYGAIN_TRACK_PEAK",
        "m4a": "----:com.apple.iTunes:REPLAYGAIN_TRACK_PEAK",
    },
    "track_replay_gain": {
        "flac": "REPLAYGAIN_TRACK_GAIN",
        "m4a": "----:com.apple.iTunes:REPLAYGAIN_TRACK_GAIN",
    },
    "title": {"flac": "TITLE", "m4a": "\xa9nam"},
}



import base64
from dataclasses import dataclass, field
from datetime import datetime, timedelta, timezone
from email.utils import parsedate_to_datetime
import json
import logging
from pathlib import Path
import sys
import time
from typing import Dict, Optional, Union

import dataclass_wizard
from platformdirs import user_config_path
import requests

PROJECT_NAME: str = "tidal-wave"
TOKEN_DIR_PATH: Path = user_config_path() / PROJECT_NAME
TOKEN_DIR_PATH.mkdir(exist_ok=True, parents=True)
OAUTH2_URL: str = "https://auth.tidal.com/v1/oauth2"
OAUTH2_HEADERS: Dict[str, str] = {
    "User-Agent": "TIDAL_ANDROID/2.38.0",
    "Accept": "application/json",
}

logger = logging.getLogger(__name__)


class AuthorizationException(Exception):
    pass


class TokenException(Exception):
    pass


@dataclass
class DeviceAuthorizationEndpointResponseJSON(dataclass_wizard.JSONSerializable):
    """This class models Tidal OAuth 2.0 API
    /device_authorization endpoint JSON response."""

    device_code: str
    user_code: str
    verification_uri_complete: str
    expires_in: int
    interval: int


@dataclass
class User:
    user_id: int
    email: Optional[str]
    country_code: str
    full_name: Optional[str]
    first_name: Optional[str]
    last_name: Optional[str]
    nickname: Optional[str]
    username: str
    address: Optional[str]
    city: Optional[str]
    postalcode: Optional[str]
    us_state: Optional[str]
    phone_number: Optional[str]
    birthday: Optional[str]
    channel_id: int
    parent_id: int
    accepted_eula: bool
    created: int
    updated: int
    facebook_uid: int
    apple_uid: Optional[str]
    google_uid: Optional[str]
    account_link_created: bool
    email_verified: bool
    new_user: bool


@dataclass
class TokenEndpointResponseJSON(dataclass_wizard.JSONSerializable):
    """This class models the JSON response from the Tidal API
    authorization /token endpoint."""

    scope: str
    user: "User"
    client_name: str
    token_type: str
    access_token: str
    refresh_token: str
    expires_in: int
    user_id: int

    def __post_init__(self):
        # Shave off 5 minutes from the expiration time. The API usually
        # gives 1-week expiration timeline, but depending on network latency
        # etc., want to refresh early.
        _timedelta = timedelta(seconds=self.expires_in - 300)
        self.expiration: datetime = datetime.now(tz=timezone.utc) + _timedelta


@dataclass
class BearerToken:
    """This class represents a token used in bearer authentication
    (https://swagger.io/docs/specification/authentication/bearer-authentication/)
    with the Tidal API."""

    access_token: str = field(repr=False)
    client_name: str  # "TIDAL_Android_2.38.0_Fire_TV_Atmos"
    expiration: Union[str, datetime] = field(repr=False)
    refresh_token: str = field(repr=False)
    user_id: int
    user_name: str

    def __post_init__(self):
        self.client_id, self.client_secret = (
            TidalOauth().client_id,
            TidalOauth().client_secret,
        )
        if isinstance(self.expiration, str):
            try:
                self.expiration = datetime.fromisoformat(self.expiration)
            except ValueError:
                raise TokenException(
                    "Expiration must be a datetime or datetime-like str"
                )

    @property
    def is_expired(self) -> bool:
        """Returns whether self.expiration is in the past, by comparing with
        datetime.datetime.now(tz=datetime.timezone.utc)."""
        return False if datetime.now(tz=timezone.utc) < self.expiration else True

    def save(self, p: Path = TOKEN_DIR_PATH / "fire_tv-tidal.token"):
        """Write some attributes as base64-encoded JSON to path on disk, p"""
        d: Dict[str, str] = {
            "access_token": self.access_token,
            "client_name": self.client_name,
            "expiration": self.expiration.isoformat(),
            "refresh_token": self.refresh_token,
            "user_id": self.user_id,
            "user_name": self.user_name,
        }
        outdata: bytes = base64.b64encode(json.dumps(d).encode("UTF-8"))
        p.write_bytes(outdata)

    @classmethod
    def load(
        cls, p: Path = TOKEN_DIR_PATH / "fire_tv-tidal.token"
    ) -> Optional["BearerToken"]:
        """Read base64-encoded JSON object from disk. If no error arises,
        return a BearerToken instance; else, return None"""

        try:
            data = json.loads(base64.b64decode(p.read_bytes()))
        except FileNotFoundError:
            logger.exception(
                TokenException(f"File '{str(p.absolute())}' does not exist")
            )
            return
        except json.JSONDecodeError:
            logger.exception(
                TokenException(f"Could not parse JSON data from '{str(p.absolute())}'")
            )
            return
        except UnicodeDecodeError:
            logger.exception(
                TokenException(
                    f"File '{str(p.absolute())}' does not appear to be base64-encoded"
                )
            )
            return

        data_args = (
            data.get(a)
            for a in (
                "access_token",
                "client_name",
                "expiration",
                "refresh_token",
                "user_id",
                "user_name",
            )
        )

        _token: BearerToken = cls(*data_args)
        return _token

    def refresh(self):
        """If self.access_token is expired, go through the token refresh process:
        https://oauth.net/2/refresh-tokens/. If successful, various attributes of
        self are overwritten: most importantly, self.expiration & self.access_token
        """
        _data = {
            "client_id": self.client_id,
            "refresh_token": self.refresh_token,
            "grant_type": "refresh_token",
            "scope": "r_usr+w_usr+w_sub",
        }
        _auth = (self.client_id, self.client_secret)
        with requests.post(
            url=f"{OAUTH2_URL}/token", data=_data, auth=_auth, headers=OAUTH2_HEADERS
        ) as resp:
            try:
                resp.raise_for_status()
            except requests.HTTPError:
                raise TokenException(
                    f"Could not refresh bearer token: HTTP error code {resp.status_code}"
                )
            else:
                token_json = resp.json()

            self.access_token = token_json.get("access_token")
            if token_json.get("clientName", token_json.get("client_name")) is not None:
                self.client_name = token_json.get(
                    "clientName", token_json.get("client_name")
                )
            if token_json.get("userId", token_json.get("user_id")) is not None:
                self.user_id = token_json.get("userId", token_json.get("user_id"))
            if token_json.get("userName", token_json.get("user_name")) is not None:
                self.user_name = token_json.get("userName", token_json.get("user_name"))

            _timedelta = timedelta(seconds=token_json.get("expires_in") - 300)
            self.expiration = datetime.now(tz=timezone.utc) + _timedelta


@dataclass
class TidalOauth:
    """This class encapsulates attributes and methods to do with authenticating
    with the Tidal OAuth API. In particular, the authorization_code_flow()
    method implements the authorization code flow part of the OAuth 2.0
    specification:
    https://auth0.com/docs/get-started/authentication-and-authorization-flow/authorization-code-flow
    The client_id and client_secret attributes are gleaned from other projects'
    work, especially
    https://github.com/Dniel97/RedSea/blob/4ba02b88cee33aeb735725cb854be6c66ff372d4/config/settings.example.py#L68
    """

    def __post_init__(self):
        self._client_id: str = "7m7Ap0JC9j1cOM3n"
        self._client_secret: str = "vRAdA108tlvkJpTsGZS8rGZ7xTlbJ0qaZ2K9saEzsgY="
        self.token: Optional[BearerToken] = None
        self.verification_url: Optional[str] = None

    @property
    def client_id(self) -> str:
        return self._client_id

    @property
    def client_secret(self) -> str:
        return self._client_secret

    def post_device_authorization(self, headers: Dict[str, str] = OAUTH2_HEADERS):
        """Send a POST request to the /device_authorization endpoint of Tidal's
        authentication API. If error, raises AuthorizationException. Else,
        return an DeviceAuthorizationEndpointResponseJSON instance with five
        attributes:
        device_code, user_code, verification_uri_complete, expires_in, interval
        """
        _url: str = f"{OAUTH2_URL}/device_authorization"
        _data: Dict[str, str] = {
            "client_id": self.client_id,
            "scope": "r_usr+w_usr+w_sub",
        }
        with requests.post(url=_url, data=_data, headers=headers) as resp:
            try:
                resp.raise_for_status()
            except requests.HTTPError as he:
                raise AuthorizationException(he.args[0])

            daerj = DeviceAuthorizationEndpointResponseJSON.from_dict(resp.json())

        self.device_authorization = daerj
        self.verification_url: str = f"http://{daerj.verification_uri_complete}"
        # "Date" header is in the "%a, %d %b %Y %H:%M:%S %Z" format:
        # e.g. "Wed, 06 Dec 2023 05:11:11 GMT".
        # So, parsedate_to_datetime converts the above into
        # datetime.datetime(2023, 12, 6, 5, 11, 11, tzinfo=datetime.timezone.utc)
        self.verification_expiration: datetime = parsedate_to_datetime(
            resp.headers.get("Date")
        ) + timedelta(seconds=daerj.expires_in)

    def authorization_code_flow(
        self, headers: Dict[str, str] = OAUTH2_HEADERS
    ) -> BearerToken:
        """Returns an instance of BearerToken by authenticating
        with the Tidal OAuth 2.0 API /token endpoint. Upon error,
        raises AuthorizationException"""
        _data = {
            "client_id": self.client_id,
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            "scope": "r_usr+w_usr+w_sub",
        }
        if self.verification_url is None:
            self.post_device_authorization()

        _data["device_code"] = self.device_authorization.device_code
        _auth = (self.client_id, self.client_secret)

        print(
            "\nCopy this URL, then navigate to it in a browser: "
            f"{self.verification_url}\n",
            file=sys.stderr,
        )

        while datetime.now(tz=timezone.utc) < self.verification_expiration:
            with requests.post(
                url=f"{OAUTH2_URL}/token", headers=headers, data=_data, auth=_auth
            ) as resp:
                if not resp.ok:
                    time.sleep(self.device_authorization.interval * 2)
                    continue
                else:
                    break
        else:
            raise AuthorizationException(
                "OAuth login process has timed out. Please try again."
            )

        logger.info("Successfully authenticated with Tidal API.")

        _token = TokenEndpointResponseJSON.from_dict(resp.json())
        if _token.token_type.lower() == "bearer":
            return BearerToken(
                access_token=_token.access_token,
                client_name=_token.client_name,
                expiration=_token.expiration,
                refresh_token=_token.refresh_token,
                user_id=_token.user_id,
                user_name=_token.user.username,
            )
        else:
            raise TokenException(
                f"Expected a bearer token, but received token type: {_token.token_type}"
            )



from dataclasses import dataclass
import json
import logging
import math
from pathlib import Path
import shutil
import sys
from types import SimpleNamespace
from typing import Dict, List, Optional, Set, Tuple, Union

import ffmpeg
import mutagen
from requests import HTTPError, Session

from .media import AudioFormat
from .models import (
    PlaylistsEndpointResponseJSON,
    TracksEndpointResponseJSON,
    VideosEndpointResponseJSON,
)
from .requesting import request_playlists
from .track import Track
from .utils import download_cover_image, temporary_file, TIDAL_API_URL
from .video import Video

logger = logging.getLogger("__name__")


@dataclass
class Playlist:
    playlist_id: str  # UUID4

    def __post_init__(self):
        self.playlist_dir: Optional[Path] = None
        self.playlist_cover_saved: bool = False

    def get_metadata(self, session: Session):
        """Request from TIDAL API /playlists endpoint"""
        self.metadata: Optional[PlaylistsEndpointResponseJSON] = request_playlists(
            session=session, identifier=self.playlist_id
        )
        self.name = (
            self.metadata.title.replace("/", "_")
            .replace("|", "_")
            .replace(":", " -")
            .replace('"', "")
            .replace("..", "")
        )

    def set_items(self, session: Session):
        """Uses data from TIDAL API /playlists/items endpoint to
        populate self.items"""
        playlist_items: Optional[PlaylistsItemsResponseJSON] = get_playlist(
            session=session, playlist_id=self.playlist_id
        )
        if playlist_items is None:
            self.items = tuple()
        else:
            self.items: Tuple[Optional[PlaylistItem]] = tuple(playlist_items.items)

    def set_dir(self, out_dir: Path):
        """Populates self.playlist_dir based on self.name, self.playlist_id"""
        playlist_substring: str = f"{self.name} [{self.playlist_id}]"
        self.playlist_dir: Path = out_dir / "Playlists" / playlist_substring
        self.playlist_dir.mkdir(parents=True, exist_ok=True)

    def save_cover_image(self, session: Session, out_dir: Path):
        """Requests self.metadata.image and attempts to write it to disk"""
        if self.playlist_dir is None:
            self.set_dir(out_dir=out_dir)
        self.cover_path: Path = self.playlist_dir / "cover.jpg"
        if not self.cover_path.exists():
            download_cover_image(
                session=session,
                cover_uuid=self.metadata.square_image,
                output_dir=self.playlist_dir,
                dimension=1080,
            )
        else:
            self.playlist_cover_saved = True

    def save_description(self):
        """Requests self.metadata.description and attempts to write it to disk"""
        description_path: Path = self.playlist_dir / "PlaylistDescription.txt"
        if self.metadata.description is not None and len(self.metadata.description) > 0:
            if not description_path.exists():
                description_path.write_text(f"{self.metadata.description}\n")

    def get_items(self, session: Session, audio_format: AudioFormat):
        """Using either Track.get() or Video.get(), attempt to request
        the data for each track or video in self.items"""
        if len(self.items) == 0:
            return
        tracks_videos: list = [None] * len(self.items)
        for i, item in enumerate(self.items):
            if item is None:
                tracks_videos[i] = None
                continue
            elif isinstance(item, TracksEndpointResponseJSON):
                track: Track = Track(track_id=item.id)
                track.get(
                    session=session,
                    audio_format=audio_format,
                    out_dir=self.playlist_dir,
                    metadata=item,
                )
                tracks_videos[i] = track
            elif isinstance(item, VideosEndpointResponseJSON):
                video: Video = Video(video_id=item.id)
                video.get(
                    session=session,
                    out_dir=self.playlist_dir,
                    metadata=item,
                )
                tracks_videos[i] = video
            else:
                tracks_videos[i] = None
                continue
        else:
            self.tracks_videos: Tuple[
                Tuple[int, Optional[Union[Track, Video]]]
            ] = tuple(tracks_videos)
        return tracks_videos

    def flatten_playlist_dir(self):
        """When self.get_items() is called, the tracks and/or videos in
        self.items are downloaded using their self-contained .get() logic;
        this means that they will be downloaded to albums. This function
        "flattens" self.playlist_dir, meaning that it moves all downloaded
        audio and video files to self.playlist_dir, and removes the various
        subdirectories created"""
        files: List[Dict[int, Optional[str]]] = [None] * len(self.tracks_videos)
        if len(self.tracks_videos) == 0:
            return
        subdirs: Set[Path] = set()

        for i, tv in enumerate(self.tracks_videos, 1):
            if getattr(tv, "outfile") is None:
                try:
                    getattr(tv, "album_dir")
                except AttributeError:
                    pass
                else:
                    subdirs.add(tv.album_dir)
                    subdirs.add(tv.album_dir.parent)
                files[i - 1] = {i: None}
                continue

            _path: Optional[Path] = Path(tv.outfile) if tv is not None else None
            # if the item never got turned into a track or video
            if _path is None:
                files[i - 1] = {i: None}
                continue

            # if the track or video didn't download
            if _path.exists():
                if _path.stat().st_size == 0:
                    files[i - 1] = {i: None}
                    continue
            else:
                files[i - 1] = {i: None}
                continue

            # otherwise, move files and clean up
            if isinstance(tv, Track):
                new_path: Path = self.playlist_dir / f"{i:03d} - {tv.trackname}"
                new_path.write_bytes(_path.read_bytes())
                _path.unlink()
                files[i - 1] = {i: str(new_path.absolute())}
            elif isinstance(tv, Video):
                new_path: Path = self.playlist_dir / f"{i:03d} - {_path.name}"
                new_path.write_bytes(_path.read_bytes())
                _path.unlink()
                files[i - 1] = {i: str(new_path.absolute())}
        else:
            self.files: List[Dict[int, Optional[str]]] = files

        # Find all subdirectories written to
        subdirs: Set[Path] = set()
        for tv in self.tracks_videos:
            if isinstance(tv, Track):
                try:
                    getattr(tv, "album_dir")
                except AttributeError:
                    pass
                else:
                    subdirs.add(tv.album_dir)
                    subdirs.add(tv.album_dir.parent)
            elif isinstance(tv, Video):
                subdirs.add(tv.artist_dir)

        # Copy all artist images, artist bio JSON files out
        # of subdirs
        artist_images: Set[Path] = set()
        for subdir in subdirs:
            for p in subdir.glob("*.jpg"):
                if p.name == "cover.jpg":
                    continue
                artist_images.add(p)
        else:
            for artist_image_path in artist_images:
                if artist_image_path.exists():
                    shutil.copyfile(
                        artist_image_path.absolute(),
                        self.playlist_dir / artist_image_path.name,
                    )

        artist_bios: Set[Path] = set()
        for subdir in subdirs:
            for p in subdir.glob("*bio.json"):
                artist_bios.add(p)
        else:
            for artist_bio_path in artist_bios:
                if artist_bio_path.exists():
                    shutil.copyfile(
                        artist_bio_path.absolute(),
                        self.playlist_dir / artist_bio_path.name,
                    )

        # Remove all subdirs
        for subdir in subdirs:
            if subdir.exists():
                shutil.rmtree(subdir)
        else:
            return self.playlist_dir

    def craft_m3u8_text(self):
        """This method creates a file called playlist.m3u8 in self.playlist_dir
        that is a standard M3U. Needs to be called after self.flatten_playlist_dir
        in order to be able to access self.files
        N.b. the already-written file is temporarily copied to a .mp4 version in a
        temporary directory because .m4a files cannot be read with mutagen."""
        m3u_text: str = f"#EXTM3U\n#EXTENC:UTF-8\n#EXTIMG:{str(self.cover_path.absolute())}\n#PLAYLIST:{self.name}\n"

        logger.info(
            f"Creating .m3u8 playlist file for Playlist with ID '{self.playlist_id}'"
        )
        for d in self.files:
            file: str = next(iter(d.values()))
            if file is None:
                continue
            elif file.endswith(".flac"):
                m = mutagen.File(file)
                artist: str = m.get("artist", [""])[0]
                title: str = m.get("title", [""])[0]
                extinf: str = (
                    f"#EXTINF:{math.ceil(m.info.length)},"
                    f"{artist} - {title}\n{file}\n"
                )
                m3u_text += extinf
            elif file.endswith(".mka"):
                m = mutagen.File(file)
                artist: str = m.get("ARTI", [""])[0]
                title: str = m.get("TITL", [""])[0]
                extinf: str = (
                    f"#EXTINF:{math.ceil(m.info.length)},"
                    f"{artist} - {title}\n{file}\n"
                )
                m3u_text += extinf
            elif file.endswith(".m4a"):
                # Mutagen cannot read .m4a files, so make a copy with all
                # of the metadata tags as a .mp4 in a temporary directory
                with temporary_file(suffix=".mp4") as tf:
                    ffmpeg.input(file, hide_banner=None, y=None).output(
                        tf.name,
                        acodec="copy",
                        vcodec="copy",
                        loglevel="quiet",
                    ).run()

                    m = mutagen.File(tf.name)
                    artist: str = m.get("\xa9ART", [""])[0]
                    title: str = m.get("\xa9nam", [""])[0]
                    extinf: str = (
                        f"#EXTINF:{math.ceil(m.info.length)},"
                        f"{artist} - {title}\n{file}\n"
                    )
                    m3u_text += extinf
        else:
            return m3u_text

    def dumps(self):
        return json.dumps(self.files)

    def dump(self, fp=sys.stdout):
        json.dump(self.files, fp)

    def get(self, session: Session, audio_format: AudioFormat, out_dir: Path):
        """The main method of this class, executing a number of other methods
        in a row:
          - self.get_metadata()
          - self.set_items()
          - self.set_dir()
          - self.save_cover_image()
          - self.save_description()
          - self.get_items()
          - self.flatten_playlist_dir()
        """
        self.get_metadata(session)
        self.set_items(session)
        self.set_dir(out_dir)
        self.save_cover_image(session, out_dir)
        try:
            self.save_description()
        except Exception:
            pass

        _get_items = self.get_items(session, audio_format)
        if _get_items is None:
            logger.critical(f"Could not retrieve playlist with ID '{self.playlist_id}'")
            return

        self.flatten_playlist_dir()

        try:
            m3u8_text: str = self.craft_m3u8_text()
        except Exception as e:
            logger.warning(
                "Unable to create playlist.m3u8 file for "
                f"playlist with ID '{self.playlist_id}'"
            )
            logger.debug(e)
        else:
            with open(self.playlist_dir / "playlist.m3u8", "w") as f:
                f.write(m3u8_text)

        logger.info(f"Playlist files written to '{self.playlist_dir}'")


class TidalPlaylistException(Exception):
    pass


def request_playlist_items(session: Session, playlist_id: str) -> Optional[dict]:
    """Request from TIDAL API /playlists/items endpoint."""
    url: str = f"{TIDAL_API_URL}/playlists/{playlist_id}/items"
    kwargs: dict = {"url": url}
    kwargs["params"] = {"limit": 100}
    kwargs["headers"] = {"Accept": "application/json"}

    data: Optional[dict] = None
    logger.info(f"Requesting from TIDAL API: playlists/{playlist_id}/items")
    with session.get(**kwargs) as resp:
        try:
            resp.raise_for_status()
        except HTTPError as he:
            if resp.status_code == 404:
                logger.warning(
                    f"404 Client Error: not found for TIDAL API endpoint playlists/{playlist_id}/items"
                )
            else:
                logger.exception(he)
        else:
            data = resp.json()
            logger.debug(
                f"{resp.status_code} response from TIDAL API to request: playlists/{playlist_id}/items"
            )
        finally:
            return data


@dataclass(frozen=True)
class PlaylistsItemsResponseJSON:
    """The response from the TIDAL API endpoint /playlists/<ID>/items
    is modeled by this class."""

    limit: int
    offset: int
    total_number_of_items: int
    items: Tuple[
        Optional[Union["TracksEndpointResponseJSON", "VideosEndpointResponseJSON"]]
    ]


def playlist_maker(
    playlists_response: Dict[str, Union[int, List[dict]]]
) -> "PlaylistsItemsResponseJSON":
    """This function massages the response from the TIDAL API endpoint
    playlists/items into a format that PlaylistsItemsResponseJSON.from_dict()
    can ingest. Returns a PlaylistsItemsResponseJSON instance"""
    init_args: dict = {}
    init_args["limit"] = playlists_response.get("limit")
    init_args["offset"] = playlists_response.get("offset")
    init_args["total_number_of_items"] = playlists_response.get("totalNumberOfItems")

    items: Tuple[SimpleNamespace] = tuple(
        SimpleNamespace(**d) for d in playlists_response["items"]
    )
    if len(items) == 0:
        return

    playlist_items: List[
        Optional[Union["TracksEndpointResponseJSON", "VideosEndpointResponseJSON"]]
    ] = [None] * init_args["total_number_of_items"]

    for i, namespace in enumerate(items):
        if namespace.type == "track":
            try:
                playlist_item = TracksEndpointResponseJSON.from_dict(namespace.item)
            except Exception as e:
                logger.warning(
                    f"TidalPlaylistException: unable to parse playlist item [i] "
                    f"with type '{namespace.type}'"
                )
                logger.debug(e)
                # value stays None
            else:
                playlist_items[i] = playlist_item
        elif namespace.type == "video":
            try:
                playlist_item = VideosEndpointResponseJSON.from_dict(namespace.item)
            except Exception as e:
                logger.warning(
                    f"TidalPlaylistException: unable to parse playlist item [i] "
                    f"with type '{namespace.type}'"
                )
                logger.debug(e)
                # value stays None
            else:
                playlist_items[i] = playlist_item
        else:
            continue  # value stays None
    else:
        init_args["items"] = tuple(playlist_items)

    return PlaylistsItemsResponseJSON(**init_args)


def get_playlist(
    session: Session, playlist_id: str
) -> Optional["PlaylistsItemsResponseJSON"]:
    """The pattern for playlist items retrieval does not follow the
    requesting.request_* functions, hence its implementation here.
    """
    playlists_items_response_json: Optional["PlaylistsItemsResponseJSON"] = None
    try:
        playlists_response: dict = request_playlist_items(
            session=session, playlist_id=playlist_id
        )
        playlists_items_response_json: Optional[
            "PlaylistsItemsResponseJSON"
        ] = playlist_maker(playlists_response=playlists_response)
    except Exception as e:
        logger.exception(TidalPlaylistException(e.args[0]))
    finally:
        return playlists_items_response_json


# union type for type hinting
PlaylistItem = Optional[
    Union["TracksEndpointResponseJSON", "VideosEndpointResponseJSON"]
]



from dataclasses import dataclass
import json
from pathlib import Path
import sys
from typing import List, Optional

from requests import Session

from .media import AudioFormat
from .models import (
    AlbumsEndpointResponseJSON,
    AlbumsItemsResponseJSON,
    AlbumsReviewResponseJSON,
)
from .requesting import request_albums, request_album_items, request_album_review
from .track import Track
from .utils import download_cover_image


@dataclass
class Album:
    album_id: int

    def __post_init__(self):
        self.album_dir: Optional[Path] = None
        self.album_cover_saved: bool = False

    def get_items(self, session: Session):
        """This method populates self.tracks by requesting from
        TIDAL albums/items endpoint."""
        album_items: AlbumsItemsResponseJSON = request_album_items(
            session=session, identifier=self.album_id
        )
        _items = album_items.items if album_items is not None else ()
        self.tracks = tuple(_item.item for _item in _items)

    def get_metadata(self, session: Session):
        """This method populates self.metadata by requesting from
        TIDAL /albums endpoint"""
        self.metadata: AlbumsEndpointResponseJSON = request_albums(
            session=session, identifier=self.album_id
        )

    def get_review(self, session: Session):
        """This method requests the review corresponding to self.album_id
        in TIDAL. If it exists, it is written to disk as AlbumReview.json
        in self.album_dir"""
        self.album_review: Optional[AlbumsReviewResponseJSON] = request_album_review(
            session=session, identifier=self.album_id
        )
        if self.album_review is not None:
            (self.album_dir / "AlbumReview.json").write_text(
                self.album_review.to_json()
            )

    def set_dir(self, out_dir: Path):
        """This method populates self.album_dir as a sub-subdirectory of
        out_dir: its parent directory is the name of the (main) artist of
        the album"""
        artist_substring: str = self.metadata.artist.name.replace("..", "")
        album_substring: str = (
            f"{self.metadata.name.replace('..', '')} "
            f"[{self.metadata.id}] [{self.metadata.release_date.year}]"
        )
        self.album_dir = out_dir / artist_substring / album_substring
        self.album_dir.mkdir(parents=True, exist_ok=True)

        if self.metadata.number_of_volumes > 1:
            for v in range(1, self.metadata.number_of_volumes + 1):
                volume_substring: str = f"Volume {v}"
                (out_dir / artist_substring / album_substring / volume_substring).mkdir(
                    parents=True, exist_ok=True
                )

    def save_cover_image(self, session: Session, out_dir: Path):
        """This method writes cover.jpg in self.album_dir via the 
        utils.download_cover_image() function. If successful,
        then self.album_cover_saved takes the value True"""
        if self.album_dir is None:
            self.set_dir(out_dir=out_dir)
        self.cover_path: Path = self.album_dir / "cover.jpg"
        if not self.cover_path.exists():
            download_cover_image(
                session=session,
                cover_uuid=self.metadata.cover,
                output_dir=self.album_dir,
            )
        else:
            self.album_cover_saved = True

    def get_tracks(
        self, session: Session, audio_format: AudioFormat, out_dir: Path
    ) -> List[Optional[str]]:
        """This method uses self.tracks to call track.Track.get() for each
        track in self.tracks. It uses the result of each of these calls to
        populate self.track_files"""
        track_files: List[str] = [None] * self.metadata.number_of_tracks
        for i, t in enumerate(self.tracks):  # type(t) is TracksEndpointResponseJSON
            track: Track = Track(track_id=t.id)

            track_files_value: Optional[str] = track.get(
                session=session,
                audio_format=audio_format,
                out_dir=out_dir,
                metadata=t,
                album=self.metadata,
            )
            track_files[i] = {track.metadata.track_number: track_files_value}
        else:
            self.track_files = track_files

    def dumps(self):
        """This method returns a JSON-like string of self.track_files"""
        return json.dumps(self.track_files)

    def dump(self, fp=sys.stdout):
        """This method writes to (by default) STDOUT a
        JSON-like string of self.track_files"""
        json.dump(self.track_files, fp)

    def get(
        self,
        session: Session,
        audio_format: AudioFormat,
        out_dir: Path,
        metadata: Optional[AlbumsEndpointResponseJSON] = None,
    ):
        """This method is the driver method of the class. It calls the
        other methods in order:
            1. get_metadata()  
            2. get_items()
            3. save_cover_image()
            4. get_review()
            5. get_tracks()
        """
        if metadata is None:
            self.get_metadata(session)
        else:
            self.metadata = metadata

        self.get_items(session)
        self.save_cover_image(session, out_dir)
        self.get_review(session)
        self.get_tracks(session, audio_format, out_dir)



from dataclasses import dataclass
import json
import logging
from pathlib import Path
import re
import shlex
import shutil
import subprocess
import sys
from typing import Dict, Iterable, List, Optional

import mutagen
from mutagen.mp4 import MP4Cover
import ffmpeg
from requests import Session

from .dash import manifester, JSONDASHManifest, Manifest, XMLDASHManifest
from .media import af_aq, AudioFormat, TAG_MAPPING
from .models import (
    AlbumsEndpointResponseJSON,
    ArtistsBioResponseJSON,
    TracksCreditsResponseJSON,
    TracksEndpointResponseJSON,
    TracksEndpointStreamResponseJSON,
    TracksLyricsResponseJSON,
)
from .requesting import (
    fetch_content_length,
    http_request_range_headers,
    request_albums,
    request_artist_bio,
    request_credits,
    request_lyrics,
    request_stream,
    request_tracks,
)
from .utils import download_artist_image, download_cover_image, temporary_file

logger = logging.getLogger("__name__")


@dataclass
class Track:
    track_id: int

    def __post_init__(self):
        self._has_lyrics: Optional[bool] = None
        self.tags: dict = {}
        self.album_cover_saved: bool = False

    def get_metadata(self, session: Session):
        self.metadata: Optional[TracksEndpointResponseJSON] = request_tracks(
            session, self.track_id
        )

    def get_album(self, session: Session):
        self.album: Optional[AlbumsEndpointResponseJSON] = request_albums(
            session, self.metadata.album.id
        )

    def get_credits(self, session: Session):
        self.credits: Optional[TracksCreditsResponseJSON] = request_credits(
            session, self.track_id
        )

    def get_lyrics(self, session: Session):
        if self._has_lyrics is None:
            self.lyrics: Optional[TracksLyricsResponseJSON] = request_lyrics(
                session, self.track_id
            )
            if self.lyrics is None:
                self._has_lyrics = False
            else:
                self._has_lyrics = True
        else:
            return self.lyrics

    def get_stream(self, session: Session, audio_format: AudioFormat):
        """Populates self.stream, self.manifest"""
        aq: Optional[str] = af_aq.get(audio_format)
        self.stream: Optional[TracksEndpointStreamResponseJSON] = request_stream(
            session, self.track_id, aq
        )

    def set_manifest(self):
        """This method sets self.manifest and self.codec"""
        self.manifest: Manifest = manifester(self.stream)
        # https://dashif.org/codecs/audio/
        if self.manifest.codecs == "flac":
            self.codec = "flac"
        elif self.manifest.codecs == "mqa":
            self.codec = "flac"
        elif self.manifest.codecs == "mha1":  # Sony 360 Reality Audio
            self.codec = "mka"
        elif self.manifest.codecs == "mp4a.40.5":  # HE-AAC
            self.codec = "m4a"
        elif self.manifest.codecs == "mp4a.40.29":  # HE-AAC v2
            self.codec = "m4a"
        elif self.manifest.codecs == "mp4a.40.2":  # AAC-LC
            self.codec = "m4a"
        elif self.manifest.codecs == "eac3":  # Enhanced AC-3
            self.codec = "m4a"
        elif self.manifest.codecs == "mp4a.40.34":  # MP3
            self.codec = "mp3"

    def set_album_dir(self, out_dir: Path):
        """This method sets self.album_dir, based on self.album and
        out_dir. In particular, self.album_dir is a subdirectory of out_dir
        based on the name of the album's artist"""
        artist_substring: str = self.album.artist.name.replace("..", "")
        album_substring: str = (
            f"{self.album.name} " f"[{self.album.id}] [{self.album.release_date.year}]"
        )
        self.album_dir: Path = out_dir / artist_substring / album_substring
        self.album_dir.mkdir(parents=True, exist_ok=True)

        if self.album.number_of_volumes > 1:
            volume_substring: str = f"Volume {self.metadata.volume_number}"
            (self.album_dir / volume_substring).mkdir(parents=True, exist_ok=True)

    def set_filename(self, audio_format: AudioFormat):
        """This method sets self.filename. It's based on self.metadata
        as well as audio_format. Additionally, if the available codecs in
        self.manifest don't match audio_format, warnings are logged"""
        _track_part: str = f"{self.metadata.track_number:02d} - {self.metadata.name}"
        if audio_format == AudioFormat.low:
            track_substring: str = f"{_track_part} [L]"
        elif audio_format == AudioFormat.high:
            track_substring: str = f"{_track_part} [H]"
        elif audio_format == AudioFormat.lossless:
            track_substring: str = f"{_track_part} [CD]"
        elif audio_format == AudioFormat.mqa:
            track_substring: str = f"{_track_part} [Q]"
        elif audio_format == AudioFormat.hi_res:
            track_substring: str = f"{_track_part} [HiRes]"
        elif audio_format == AudioFormat.dolby_atmos:
            track_substring: str = f"{_track_part} [A]"
        elif audio_format == AudioFormat.sony_360_reality_audio:
            track_substring: str = f"{_track_part} [360]"
        else:
            track_substring: str = _track_part

        # Check for MQA masquerading as HiRes here
        if audio_format == AudioFormat.hi_res:
            if self.manifest.codecs == "mqa":
                logger.warning(
                    "Even though HiRes audio format was requested, this track is only "
                    "available in MQA format. TIDAL regards this as 'HiRes' even though "
                    "it is probably only lossless; i.e. 16-bit 44.1 kHz quality. "
                    "Downloading of track will continue, but it will be marked as MQA."
                )
                self.filename: Optional[str] = f"{_track_part} [Q].{self.codec}"
            elif (self.stream.bit_depth == 16) and (self.stream.sample_rate == 44100):
                logger.warning(
                    "Even though HiRes audio format was requested, and TIDAL responded to "
                    "that request without error, this track is only available in lossless "
                    "format; i.e. 16-bit 44.1 kHz quality. Downloading of track will "
                    "continue, but it will be marked as Lossless ([CD])."
                )
                self.filename: Optional[str] = f"{_track_part} [CD].{self.codec}"
            else:
                self.filename: Optional[str] = f"{track_substring}.{self.codec}"
        else:
            self.filename: Optional[str] = f"{track_substring}.{self.codec}"

        # for use in playlist file ordering
        self.trackname: str = re.match(r"(?:\d{2,3} - )(.+?$)", self.filename).groups()[
            0
        ]

    def set_outfile(self):
        """Uses self.album_dir and self.metadata and self.filename
        to craft the pathlib.Path object, self.outfile, that is a
        reference to where the track will be written on disk."""
        if self.album.number_of_volumes > 1:
            self.outfile: Path = (
                self.album_dir / f"Volume {self.metadata.volume_number}" / self.filename
            )
            self.absolute_outfile = str(self.outfile.absolute())
        else:
            self.outfile: Path = self.album_dir / self.filename
            self.absolute_outfile = str(self.outfile.absolute())

        if (self.outfile.exists()) and (self.outfile.stat().st_size > 0):
            logger.info(
                f"Track {self.absolute_outfile} already exists "
                "and therefore will not be overwritten"
            )
            return
        else:
            return self.outfile

    def save_artist_image(self, session: Session):
        """This method writes a JPEG file with the name of each of
        self.metadata.artists to self.album_dir"""
        for a in self.metadata.artists:
            track_artist_image: Path = (
                self.album_dir / f"{a.name.replace('..', '')}.jpg"
            )
            if not track_artist_image.exists():
                download_artist_image(session, a, self.album_dir)

    def save_artist_bio(self, session: Session):
        """This method writes a JSON file with the name of each of
        self.metadata.artists to self.album_dir"""
        for a in self.metadata.artists:
            track_artist_bio_json: Path = self.album_dir / f"{a.name}-bio.json"
            if not track_artist_bio_json.exists():
                artist_bio: Optional[ArtistsBioResponseJSON] = request_artist_bio(
                    session, a.id
                )
                if artist_bio is not None:
                    logger.info(
                        f"Writing artist bio for artist {a.id} to "
                        f"'{str(track_artist_bio_json.absolute())}"
                    )
                    track_artist_bio_json.write_text(artist_bio.to_json())

    def save_album_cover(self, session: Session):
        """This method saves cover.jpg to self.album_dir; the bytes for cover.jpg
        come from self.album.cover"""
        self.cover_path: Path = self.album_dir / "cover.jpg"
        if (not self.cover_path.exists()) or (not self.album_cover_saved):
            download_cover_image(
                session=session, cover_uuid=self.album.cover, output_dir=self.album_dir
            )
        else:
            self.album_cover_saved = True

    def set_urls(self, session: Session):
        """This method sets self.urls based on self.manifest"""
        if isinstance(self.manifest, JSONDASHManifest):
            self.urls: List[str] = self.manifest.urls
        elif isinstance(self.manifest, XMLDASHManifest):
            self.urls: List[str] = self.manifest.build_urls(session=session)
        self.download_headers: Dict[str, str] = {"Accept": self.manifest.mime_type}
        if session.session_id is not None:
            self.download_headers["sessionId"] = session.session_id
        self.download_params = {k: None for k in session.params}

    def download_url(self, session: Session, out_dir: Path) -> Optional[Path]:
        """This method downloads self.urls[0], for use in situations when
        the manifest returned by TIDAL API contains one URL. It relies on
        byte range headers to incrementally get all content from a URL"""
        logger.info(f"Writing track {self.track_id} to '{self.absolute_outfile}'")

        with temporary_file() as ntf:
            # Implement HTTP range requests here to mimic official clients
            range_size: int = 1024 * 1024  # 1 MiB
            content_length: int = fetch_content_length(
                session=session, url=self.urls[0]
            )
            if content_length == 0:
                return

            range_headers: Iterable[str] = http_request_range_headers(
                content_length=content_length,
                range_size=range_size,
                return_tuple=False,
            )
            for rh in range_headers:
                with session.get(
                    self.urls[0], params=self.download_params, headers={"Range": rh}
                ) as rr:
                    if not rr.ok:
                        logger.warning(f"Could not download {self}")
                        return
                    else:
                        ntf.write(rr.content)
            else:
                ntf.seek(0)

            if self.codec == "flac":
                # Have to use FFMPEG to re-mux the audio bytes, otherwise
                # mutagen chokes on NoFlacHeaderError
                ffmpeg.input(ntf.name, hide_banner=None, y=None).output(
                    self.absolute_outfile,
                    acodec="copy",
                    loglevel="quiet",
                ).run()
            elif self.codec == "m4a":
                shutil.copyfile(ntf.name, self.outfile)
            elif self.codec == "mka":
                shutil.copyfile(ntf.name, self.outfile)

            logger.info(
                f"Track {self.track_id} written to '{str(self.outfile.absolute())}'"
            )
            return self.outfile

    def download_urls(self, session: Session, out_dir: Path) -> Optional[Path]:
        """This method writes the contents from self.urls to a temporary
        directory, then uses FFmpeg to re-mux the data to self.outfile"""
        logger.info(f"Writing track {self.track_id} to '{self.absolute_outfile}'")

        with temporary_file() as ntf:
            for u in self.urls:
                with session.get(
                    url=u, headers=self.download_headers, params=self.download_params
                ) as resp:
                    if not resp.ok:
                        logger.warning(f"Could not download {self}")
                        return
                    else:
                        ntf.write(resp.content)
            else:
                ntf.seek(0)

            if self.codec == "flac":
                # Have to use FFmpeg to re-mux the audio bytes, otherwise
                # mutagen chokes on NoFlacHeaderError
                ffmpeg.input(ntf.name, hide_banner=None, y=None).output(
                    self.absolute_outfile, acodec="copy", loglevel="quiet"
                ).run()
            elif self.codec == "m4a":
                shutil.copyfile(ntf.name, self.outfile)
            elif self.codec == "mka":
                shutil.copyfile(ntf.name, self.outfile)

        logger.info(f"Track {self.track_id} written to '{self.absolute_outfile}'")
        return self.outfile

    def download(self, session: Session, out_dir: Path) -> Optional[Path]:
        """This method GETs the data from self.urls and writes it
        to self.outfile."""
        if len(self.urls) == 1:
            outfile: Optional[Path] = self.download_url(
                session=session, out_dir=out_dir
            )
        else:
            outfile: Optional[Path] = self.download_urls(
                session=session, out_dir=out_dir
            )

        return outfile

    def craft_tags(self):
        """Using the TAG_MAPPING dictionary,
        write the correct values of various metadata tags to the file.
        E.g. for .flac files, the album's artist is 'ALBUMARTIST',
        but for .m4a files, the album's artist is 'aART'."""
        tags = dict()
        if (self.codec == "flac") or (self.codec == "mka"):
            tag_map = {k: v["flac"] for k, v in TAG_MAPPING.items()}
        elif self.codec == "m4a":
            tag_map = {k: v["m4a"] for k, v in TAG_MAPPING.items()}

        tags[tag_map["album"]] = self.album.title
        tags[tag_map["album_artist"]] = ";".join((a.name for a in self.album.artists))
        tags[tag_map["album_peak_amplitude"]] = f"{self.stream.album_peak_amplitude}"
        tags[tag_map["album_replay_gain"]] = f"{self.stream.album_replay_gain}"
        tags[tag_map["artist"]] = ";".join((a.name for a in self.metadata.artists))
        tags[tag_map["artists"]] = [a.name for a in self.metadata.artists]
        tags[tag_map["barcode"]] = self.album.upc
        tags[tag_map["comment"]] = self.metadata.url
        tags[tag_map["copyright"]] = self.metadata.copyright
        tags[tag_map["date"]] = str(self.album.release_date)
        tags[tag_map["isrc"]] = self.metadata.isrc
        tags[tag_map["title"]] = self.metadata.name
        tags[tag_map["track_peak_amplitude"]] = f"{self.metadata.peak}"
        tags[tag_map["track_replay_gain"]] = f"{self.metadata.replay_gain}"
        # credits
        for tag in {"composer", "engineer", "lyricist", "mixer", "producer", "remixer"}:
            try:
                _credits_tag = ";".join(getattr(self.credits, tag))
            except (TypeError, AttributeError):  # NoneType problems
                continue
            else:
                tags[tag_map[tag]] = _credits_tag
        # lyrics
        try:
            _lyrics = self.lyrics.subtitles
        except (TypeError, AttributeError):  # NoneType problems
            pass
        else:
            tags[tag_map["lyrics"]] = _lyrics

        if self.codec == "flac":
            # track and disk
            tags["DISCTOTAL"] = f"{self.metadata.volume_number}"
            tags["DISC"] = f"{self.album.number_of_volumes}"
            tags["TRACKTOTAL"] = f"{self.album.number_of_tracks}"
            tags["TRACKNUMBER"] = f"{self.metadata.track_number}"
            # instrument-specific
            #     piano
            try:
                piano_credits: List[str] = [
                    f"{pc} (piano)" for pc in self.credits.piano
                ]
            except (TypeError, AttributeError):  # NoneType problems
                pass
            else:
                tags["PERFORMER"] = piano_credits

        elif self.codec == "m4a":
            # Have to convert to bytes the values of the tags starting with '----'
            for k, v in tags.copy().items():
                if k.startswith("----"):
                    if isinstance(v, str):
                        tags[k]: bytes = v.encode("UTF-8")
                    elif isinstance(v, list):
                        tags[k]: List[bytes] = [s.encode("UTF-8") for s in v]

            tags["trkn"] = [(self.metadata.track_number, self.album.number_of_tracks)]
            tags["disk"] = [(self.metadata.volume_number, self.album.number_of_volumes)]

        self.tags: dict = {k: v for k, v in tags.items() if v is not None}

    def set_tags(self):
        """Instantiate a mutagen.File instance, add self.tags to it, and
        save it to disk"""
        self.mutagen = mutagen.File(self.outfile)
        self.mutagen.clear()
        self.mutagen.update(**self.tags)
        # add album cover
        if self.codec == "flac":
            p = mutagen.flac.Picture()
            p.type = mutagen.id3.PictureType.COVER_FRONT
            p.desc = "Album Cover"
            p.width = p.height = 1280
            p.mime = "image/jpeg"
            p.data = self.cover_path.read_bytes()
            self.mutagen.add_picture(p)
        elif self.codec == "m4a":
            self.mutagen["covr"] = [
                MP4Cover(self.cover_path.read_bytes(), imageformat=MP4Cover.FORMAT_JPEG)
            ]

        self.mutagen.save()
        # Make sure audio track comes first because of
        # less-sophisticated audio players that only
        # recognize the first stream
        if self.codec == "flac":
            with temporary_file(suffix=".mka") as tf:
                shutil.move(str(self.outfile.absolute()), tf.name)
                cmd: List[str] = shlex.split(
                    f"""ffmpeg -hide_banner -loglevel quiet -y -i "{tf.name}"
                    -map 0:a:0 -map 0:v:0 -c:a copy -c:v copy
                    -metadata:s:v title='Album cover' -metadata:s:v comment='Cover (front)'
                    -disposition:v attached_pic "{self.absolute_outfile}" """
                )
                subprocess.run(cmd)
        elif self.codec == "m4a":
            with temporary_file(suffix=".mka") as tf:
                cmd: List[str] = shlex.split(
                    f"""ffmpeg -hide_banner -loglevel quiet -y -i "{self.absolute_outfile}"
                    -map 0:a:0 -map 0:v:0 -c:a copy -c:v copy "{tf.name}" """
                )
                subprocess.run(cmd)
                shutil.copyfile(tf.name, self.absolute_outfile)

    def get(
        self,
        session: Session,
        audio_format: AudioFormat,
        out_dir: Path,
        metadata: Optional[TracksEndpointResponseJSON] = None,
        album: Optional[AlbumsEndpointResponseJSON] = None,
    ) -> Optional[str]:
        if metadata is None:
            self.get_metadata(session)
        else:
            self.metadata = metadata

        if self.metadata is None:
            # self.failed = True
            self.outfile = None
            return

        if "DOLBY_ATMOS" in self.metadata.media_metadata.tags:
            if audio_format != AudioFormat.dolby_atmos:
                logger.warning(
                    f"Track {self.track_id} is only available in Dolby Atmos "
                    "format. Downloading of track will not continue."
                )
                self.outfile = None
                return

        if audio_format == AudioFormat.dolby_atmos:
            if "DOLBY_ATMOS" not in self.metadata.media_metadata.tags:
                logger.warning(
                    "Dolby Atmos audio format was requested, but track "
                    f"{self.track_id} is not available in Dolby Atmos "
                    "format. Downloading of track will not continue."
                )
                self.outfile = None
                return
        elif audio_format == AudioFormat.sony_360_reality_audio:
            if "SONY_360RA" not in self.metadata.media_metadata.tags:
                logger.warning(
                    "Sony 360 Reality Audio audio format was requested, but track "
                    f"{self.track_id} is not available in Sony 360 Reality Audio "
                    "format. Downloading of track will not continue."
                )
                self.outfile = None
                return
        elif audio_format == AudioFormat.mqa:
            if "MQA" not in self.metadata.media_metadata.tags:
                logger.warning(
                    "MQA audio format was requested, but track "
                    f"{self.track_id} is not available in MQA audio "
                    "format. Downloading of track will not continue."
                )
                self.outfile = None
                return

        if album is None:
            self.get_album(session)
        else:
            self.album = album

        if self.album is None:
            self.outfile = None
            return

        self.get_credits(session)
        self.get_stream(session, audio_format)
        if self.stream is None:
            return
        self.set_manifest()
        self.set_album_dir(out_dir)
        self.set_filename(audio_format)
        outfile: Optional[Path] = self.set_outfile()
        if outfile is None:
            return

        try:
            self.get_lyrics(session)
        except Exception:
            pass

        self.save_album_cover(session)

        try:
            self.save_artist_image(session)
        except Exception:
            pass

        try:
            self.save_artist_bio(session)
        except Exception:
            pass

        self.set_urls(session)

        if self.download(session, out_dir) is None:
            return

        self.craft_tags()
        self.set_tags()

        return str(self.outfile.absolute())

    def dump(self, fp=sys.stdout):
        k: int = int(self.metadata.track_number)
        if self.outfile is None:
            v: Optional[str] = None
        elif not isinstance(self.outfile, Path):
            v: Optional[str] = None
        else:
            v: Optional[str] = str(self.outfile.absolute())
        json.dump({k: v}, fp)
        return None

    def dumps(self) -> str:
        k: int = int(self.metadata.track_number)
        if self.outfile is None:
            v: Optional[str] = None
        elif not isinstance(self.outfile, Path):
            v: Optional[str] = None
        else:
            v: Optional[str] = str(self.outfile.absolute())
        json.dumps({k: v})
        return None



from functools import partial
import logging
from typing import Callable, Iterable, Iterator, Optional, Tuple, Union

from .models import (
    AlbumsEndpointResponseJSON,
    AlbumsItemsResponseJSON,
    AlbumsReviewResponseJSON,
    ArtistsAlbumsResponseJSON,
    ArtistsBioResponseJSON,
    ArtistsEndpointResponseJSON,
    ArtistsVideosResponseJSON,
    PlaylistsEndpointResponseJSON,
    SessionsEndpointResponseJSON,
    SubscriptionEndpointResponseJSON,
    TracksCreditsResponseJSON,
    TracksEndpointResponseJSON,
    TracksEndpointStreamResponseJSON,
    TracksLyricsResponseJSON,
    VideosContributorsResponseJSON,
    VideosEndpointResponseJSON,
    VideosEndpointStreamResponseJSON,
)
from .utils import TIDAL_API_URL

import backoff
from requests import HTTPError, Response, Session

logger: logging.Logger = logging.getLogger(__name__)

ResponseJSON = Union[
    AlbumsEndpointResponseJSON,
    AlbumsItemsResponseJSON,
    AlbumsReviewResponseJSON,
    ArtistsAlbumsResponseJSON,
    ArtistsBioResponseJSON,
    ArtistsEndpointResponseJSON,
    ArtistsVideosResponseJSON,
    PlaylistsEndpointResponseJSON,
    SessionsEndpointResponseJSON,
    SubscriptionEndpointResponseJSON,
    TracksCreditsResponseJSON,
    TracksEndpointResponseJSON,
    TracksEndpointStreamResponseJSON,
    TracksLyricsResponseJSON,
    VideosContributorsResponseJSON,
    VideosEndpointResponseJSON,
    VideosEndpointStreamResponseJSON,
]


def requester_maker(
    session: Session,
    endpoint: str,
    identifier: int,
    url_end: str = "",
    headers: Optional[dict] = None,
    parameters: Optional[dict] = None,
    subclass: Optional["ResponseJSON"] = None,
    credits_flag: bool = False,
) -> Callable:
    """This function is a function factory: it crafts nearly identical
    versions of the same logic: send a GET request to a certain endpoint;
    if a requests.HTTPError arises, return None; else, transform the
    JSON response into an instance of a subclass of JSONWizard."""

    def function(s, e, i, u, h, p, sc, cf):
        url: str = f"{TIDAL_API_URL}/{e}/{i}{u}"
        kwargs: dict = {"url": url}
        if p is not None:
            kwargs["params"] = p
        if h is not None:
            kwargs["headers"] = h

        @backoff.on_predicate(
            backoff.expo,
            predicate=lambda r: r.status_code == 429,
            jitter=backoff.random_jitter,
            max_time=10,
            logger=logger,
        )
        def _get(s: Session, request_kwargs: dict) -> Response:
            """Return a requests.Response object, from having passed request_kwargs
            to s.get(), optionally retrying if 429 error occurs."""
            with s.get(**request_kwargs) as r:
                return r

        data: Optional[sc] = None
        logger.info(f"Requesting from TIDAL API: {e}/{i}{u}")
        resp: Response = _get(s=s, request_kwargs=kwargs)

        try:
            resp.raise_for_status()
        except HTTPError as he:
            if resp.status_code == 404:
                logger.warning(
                    f"404 Client Error: not found for TIDAL API endpoint {e}/{i}{u}"
                )
            elif resp.status_code == 401:
                logger.warning(
                    f"401 Client Error: Unauthorized for TIDAL API endpoint {e}/{i}{u}"
                )
            else:
                logger.exception(he)
        else:
            if cf:
                data = sc.from_dict({"credits": resp.json()})
            else:
                data = sc.from_dict(resp.json())
        finally:
            return data

    return function(
        s=session,
        e=endpoint,
        i=identifier,
        u=url_end,
        h=headers,
        p=parameters,
        sc=subclass,
        cf=credits_flag,
    )


# Functions will curry all arguments except session and identifier, as those
# don't change during runtime, and session is only available once the __main__
# process (on the happy path) creates a requests.Session object. Identifier
# varies with the media type, etc.


def request_albums(
    session: Session, identifier: int
) -> Optional[AlbumsEndpointResponseJSON]:
    return requester_maker(
        session=session,
        endpoint="albums",
        identifier=identifier,
        headers={"Accept": "application/json"},
        subclass=AlbumsEndpointResponseJSON,
    )


def request_album_items(
    session: Session, identifier: int
) -> Optional[AlbumsItemsResponseJSON]:
    return requester_maker(
        session=session,
        endpoint="albums",
        identifier=identifier,
        headers={"Accept": "application/json"},
        parameters={"limit": 100},
        url_end="/items",
        subclass=AlbumsItemsResponseJSON,
    )


def request_album_review(
    session: Session, identifier: int
) -> Optional[AlbumsReviewResponseJSON]:
    return requester_maker(
        session=session,
        endpoint="albums",
        identifier=identifier,
        headers={"Accept": "application/json"},
        url_end="/review",
        subclass=AlbumsReviewResponseJSON,
    )


def request_artist_bio(
    session: Session, identifier: int
) -> Optional[ArtistsBioResponseJSON]:
    return requester_maker(
        session=session,
        endpoint="artists",
        identifier=identifier,
        headers={"Accept": "application/json"},
        url_end="/bio",
        subclass=ArtistsBioResponseJSON,
    )


def request_artists(
    session: Session, identifier: int
) -> Optional[ArtistsEndpointResponseJSON]:
    return requester_maker(
        session=session,
        endpoint="artists",
        identifier=identifier,
        headers={"Accept": "application/json"},
        subclass=ArtistsEndpointResponseJSON,
    )


def request_artists_albums(
    session: Session, identifier: int
) -> Optional[ArtistsAlbumsResponseJSON]:
    return requester_maker(
        session=session,
        endpoint="artists",
        identifier=identifier,
        headers={"Accept": "application/json"},
        url_end="/albums",
        subclass=ArtistsAlbumsResponseJSON,
    )


def request_artists_audio_works(
    session: Session, identifier: int
) -> Optional[ArtistsAlbumsResponseJSON]:
    return requester_maker(
        session=session,
        endpoint="artists",
        identifier=identifier,
        headers={"Accept": "application/json"},
        parameters={"filter": "EPSANDSINGLES"},
        url_end="/albums",
        subclass=ArtistsAlbumsResponseJSON,
    )


def request_artists_videos(
    session: Session, identifier: int
) -> Optional[ArtistsVideosResponseJSON]:
    return requester_maker(
        session=session,
        endpoint="artists",
        identifier=identifier,
        headers={"Accept": "application/json"},
        url_end="/videos",
        subclass=ArtistsVideosResponseJSON,
    )


def request_tracks(
    session: Session, identifier: int
) -> Optional[TracksEndpointResponseJSON]:
    return requester_maker(
        session=session,
        endpoint="tracks",
        identifier=identifier,
        headers={"Accept": "application/json"},
        subclass=TracksEndpointResponseJSON,
    )


# This one's special, because its JSON response isn't proper JSON:
# it's just an array of JSON objects, so we have to pass a flag to mark
# that the logic common to the rest of the functions is slightly different here.
def request_credits(
    session: Session, identifier: int
) -> Optional[TracksCreditsResponseJSON]:
    return requester_maker(
        session=session,
        endpoint="tracks",
        identifier=identifier,
        headers={"Accept": "application/json"},
        parameters={"includeContributors": True},
        url_end="/credits",
        subclass=TracksCreditsResponseJSON,
        credits_flag=True,
    )


def request_lyrics(
    session: Session, identifier: int
) -> Optional[TracksLyricsResponseJSON]:
    return requester_maker(
        session=session,
        endpoint="tracks",
        identifier=identifier,
        headers={"Accept": "application/json"},
        url_end="/lyrics",
        subclass=TracksLyricsResponseJSON,
    )


# One more layer of currying here, as the parameters argument
# is dependent on a runtime variable.


def request_stream(
    session: Session, track_id: int, audio_quality: str
) -> Optional[TracksEndpointStreamResponseJSON]:
    func = partial(
        requester_maker,
        session=session,
        endpoint="tracks",
        identifier=track_id,
        headers={"Accept": "application/json"},
        parameters={
            "audioquality": audio_quality,
            "playbackmode": "STREAM",
            "assetpresentation": "FULL",
        },
        url_end="/playbackinfopostpaywall",
        subclass=TracksEndpointStreamResponseJSON,
    )
    return func()


def request_videos(
    session: Session, identifier: int
) -> Optional[VideosEndpointResponseJSON]:
    return requester_maker(
        session=session,
        endpoint="videos",
        identifier=identifier,
        headers={"Accept": "application/json"},
        subclass=VideosEndpointResponseJSON,
    )


def request_video_contributors(
    session: Session, identifier: int
) -> Optional[VideosContributorsResponseJSON]:
    return requester_maker(
        session=session,
        endpoint="videos",
        identifier=identifier,
        headers={"Accept": "application/json"},
        parameters={"limit": 100},
        url_end="/contributors",
        subclass=VideosContributorsResponseJSON,
    )


def request_video_stream(
    session: Session, video_id: int, video_quality: str
) -> Optional[VideosEndpointStreamResponseJSON]:
    func = partial(
        requester_maker,
        session=session,
        identifier=video_id,
        endpoint="videos",
        headers={"Accept": "application/json"},
        parameters={
            "videoquality": video_quality,
            "playbackmode": "STREAM",
            "assetpresentation": "FULL",
        },
        url_end="/playbackinfopostpaywall",
        subclass=VideosEndpointStreamResponseJSON,
    )
    return func()


def request_playlists(
    session: Session, identifier: int
) -> Optional[PlaylistsEndpointResponseJSON]:
    return requester_maker(
        session=session,
        endpoint="playlists",
        identifier=identifier,
        headers={"Accept": "application/json"},
        subclass=PlaylistsEndpointResponseJSON,
    )


def get_album_id(session: Session, track_id: int) -> Optional[int]:
    """Given the Tidal ID to a track, query the Tidal API in order to retrieve
    the Tidal ID of the album to which the track belongs"""
    terj: Optional[TracksEndpointResponseJSON] = request_tracks(session, track_id)
    album_id: Optional[int] = None

    try:
        album_id = terj.id
    except AttributeError:
        pass
    finally:
        return album_id


def contiguous_ranges(value: int, range_size: int) -> Iterator[Tuple[int, int]]:
    """This function is a generator: it yields two-tuples of int, with the
    tuples representing the (inclusive) boundaries of ranges of size
    range_size. The final tuple will represent a range <= range_size if
    range_size does not evenly divide value. E.g.
    ```>>> list(ranges(16, 3))
    [(0, 2), (3, 5), (6, 8), (9, 11), (12, 14), (15, 16)]
    ```
    N.b. the first tuple will always have first element 0 and the final tuple
    will always have second element `value`."""
    i: int = 0
    rs: int = range_size - 1
    while i + rs < value:
        t: Tuple[int, int] = (i, i + rs)
        i = t[-1] + 1
        yield t
    else:
        yield (i, value)


def http_request_range_headers(
    content_length: int, range_size: int, return_tuple: bool = True
) -> Iterable[str]:
    """This function creates HTTP request Range headers. Its iterable
    returned is of tuples; each tuple describes the (inclusive) boundaries
    of a bytes range with size range_size. If return_tuple is False, it returns
    a generator of tuples. E.g.
    ```>>> http_request_range_headers(16, 3)
    ('bytes=0-2',
     'bytes=3-5',
     'bytes=6-8',
     'bytes=9-11',
     'bytes=12-14',
     'bytes=15-16')
    ```
    """
    ranges: Iterator[Tuple[int, int]] = contiguous_ranges(content_length, range_size)
    iterable: Iterable = (f"bytes={t[0]}-{t[1]}" for t in ranges)
    if return_tuple:
        return tuple(iterable)
    else:
        return iterable


def fetch_content_length(session: Session, url: str) -> int:
    """Attempt to get the amount of bytes pointed to by `url`. If
    the HEAD request from the requests.Session object, `session`,
    encounters an HTTP request; or if the server does not support
    HTTP range requests; or if the server does not response with a
    Content-Length header, return 0"""
    session_params: dict = session.params
    # Unset params to avoid 403 response
    _params: dict = {k: None for k in session_params}
    with session.head(url=url, params=_params) as resp:
        if not resp.ok:
            cl: str = "0"
        else:
            cl: str = resp.headers.get("Content-Length", "0")
    return int(cl)



from dataclasses import dataclass
import logging
from pathlib import Path
from typing import List, Optional

from requests import Session

from .album import Album
from .media import AudioFormat
from .models import (
    ArtistsAlbumsResponseJSON,
    ArtistsEndpointResponseJSON,
    ArtistsVideosResponseJSON,
)
from .requesting import (
    request_artists,
    request_artists_albums,
    request_artists_audio_works,
    request_artists_videos,
)
from .utils import download_cover_image
from .video import Video

logger = logging.getLogger("__name__")


@dataclass
class Artist:
    artist_id: int

    def set_metadata(self, session: Session):
        """This function requests from TIDAL API endpoint /artists and
        stores the results in self.metadata"""
        self.metadata: Optional[ArtistsEndpointResponseJSON] = request_artists(
            session, self.artist_id
        )

    def save_artist_image(self, session: Session):
        """This method writes the bytes of self.metadata.picture to
        the file cover.jpg in self.artist_dir"""
        artist_image: Path = self.artist_dir / "cover.jpg"
        if not artist_image.exists():
            download_cover_image(
                session, self.metadata.picture, self.artist_dir, dimension=750
            )

    def set_albums(self, session: Session):
        """This method requests from TIDAL API endpoint /artists/albums and
        stores the results in self.albums"""
        self.albums: Optional[ArtistsAlbumsResponseJSON] = request_artists_albums(
            session, self.artist_id
        )

    def set_audio_works(self, session: Session):
        """This method requests from TIDAL API endpoint
        /artists/albums?filter=EPSANDSINGLES and stores the results in self.albums"""
        self.albums: Optional[ArtistsAlbumsResponseJSON] = request_artists_audio_works(
            session, self.artist_id
        )

    def set_videos(self, session: Session):
        """This method requests from TIDAL API endpoint /artists/videos and
        stores the results in self.albums"""
        self.videos: Optional[ArtistsVideosResponseJSON] = request_artists_videos(
            session, self.artist_id
        )

    def set_dir(self, out_dir: Path):
        """This method sets self.artist_dir and creates the directory on the file system
        if it does not exist"""
        self.name: str = self.metadata.name.replace("..", "")
        self.artist_dir = out_dir / self.name
        self.artist_dir.mkdir(parents=True, exist_ok=True)

    def get_albums(
        self,
        session: Session,
        audio_format: AudioFormat,
        out_dir: Path,
        include_eps_singles: bool = False,
    ) -> List[Optional[str]]:
        """This method first fetches the total albums on TIDAL's service
        corresponding to the artist with ID self.artist_id. Then, each of
        the albums (and, optionally, EPs and singles) is requested and
        written to subdirectories of out_dir"""
        if include_eps_singles:
            self.set_audio_works(session)
            logger.info(
                f"Starting attempt to get {self.albums.total_number_of_items} "
                "albums, EPs, and singles for artist with ID "
                f"{self.metadata.id},  '{self.name}'"
            )
        else:
            self.set_albums(session)
            logger.info(
                f"Starting attempt to get {self.albums.total_number_of_items} albums "
                f"for artist with ID {self.metadata.id}, '{self.name}'"
            )

        for i, a in enumerate(self.albums.items):
            album: Album = Album(album_id=a.id)
            album.get(
                session=session,
                audio_format=audio_format,
                out_dir=out_dir,
                metadata=a,
            )

    def get_videos(
        self,
        session: Session,
        out_dir: Path,
    ) -> List[Optional[str]]:
        """This method sets self.videos by calling self.set_videos()
        then, for each video, instantiates a Video object and executes
        video.get()"""
        self.set_videos(session)
        logger.info(
            f"Starting attempt to get {self.videos.total_number_of_items} videos "
            f"for artist with ID {self.metadata.id}, '{self.name}'"
        )
        for i, v in enumerate(self.videos.items):
            video: Video = Video(video_id=v.id)
            video.get(
                session=session,
                out_dir=out_dir,
                metadata=v,
            )

    def get(
        self,
        session: Session,
        audio_format: AudioFormat,
        out_dir: Path,
        include_eps_singles: bool,
    ):
        """This is the driver method of the class. It executes the other
        methods in order:
            1. set_metadata
            2. set_dir
            3. save_artist_image
            4. get_videos
            5. get_albums
        """
        self.set_metadata(session)
        self.set_dir(out_dir)
        self.save_artist_image(session)
        self.get_videos(session, out_dir)
        if include_eps_singles:
            self.get_albums(session, audio_format, out_dir, include_eps_singles=True)
        self.get_albums(session, audio_format, out_dir, include_eps_singles=False)



import base64
from dataclasses import dataclass, field
from datetime import date, datetime
import logging
import re
from typing import List, Literal, Optional, Tuple, Union
from typing_extensions import Annotated

import dataclass_wizard
from requests.auth import AuthBase

logger = logging.getLogger(__name__)
IMAGE_URL: str = "https://resources.tidal.com/images/%s.jpg"
AudioModeType = Literal["DOLBY_ATMOS", "SONY_360RA", "STEREO"]
AudioQualityType = Literal[
    "HI_RES", "HI_RES_LOSSLESS", "LOSSLESS", "DOLBY_ATMOS", "HIGH", "LOW"
]
VideoQualityType = Literal["HIGH", "MEDIUM", "LOW", "AUDIO_ONLY"]


@dataclass
class TracksEndpointStreamResponseJSON(dataclass_wizard.JSONWizard):
    """Response from the TIDAL API's tracks/{TRACKID} stream
    endpoint. The params and headers, if correctly specified, return the
    manifest of the audio to be streamed. The manifest is a base64-encoded
    XML document or JSON object"""

    track_id: int
    audio_mode: AudioModeType
    audio_quality: AudioQualityType
    manifest: str = field(repr=False)
    manifest_mime_type: str = field(repr=False)
    album_replay_gain: Optional[float] = field(repr=False, default=None)
    album_peak_amplitude: Optional[float] = field(repr=False, default=None)
    track_replay_gain: Optional[float] = field(repr=False, default=None)
    track_peak_amplitude: Optional[float] = field(repr=False, default=None)
    bit_depth: Optional[int] = field(default=None)
    sample_rate: Optional[int] = field(default=None)

    def __post_init__(self):
        self.manifest_bytes: bytes = base64.b64decode(self.manifest)


class BearerAuth(AuthBase):
    """A class to be passed to the `auth` argument in a `requests.Session`
    constructor"""

    def __init__(self, token: str):
        self.token = token

    def __call__(self, r):
        r.headers["Authorization"] = f"Bearer {self.token}"
        return r


@dataclass(frozen=True)
class Client:
    id: int
    name: str
    authorized_for_offline: bool = field(repr=False)


@dataclass(frozen=True)
class SessionsEndpointResponseJSON(dataclass_wizard.JSONWizard):
    session_id: str  # UUID4 value, really
    user_id: int
    country_code: str  # 2-digit country code according to some ISO
    channel_id: int
    partner_id: int
    client: "Client"


@dataclass(frozen=True)
class Artist:
    """A musical artist in the reckoning of the TIDAL API"""

    id: int
    name: str
    type: str
    picture: Optional[str] = field(repr=False, default=None)

    def picture_url(self, dimension: int = 320) -> Optional[str]:
        if self.picture is None:
            return
        elif len(self.picture) != 36 or self.picture.count("-") != 4:
            # Should be a UUID
            return
        else:
            _picture = self.picture.replace("-", "/")
            return IMAGE_URL % f"{_picture}/{dimension}x{dimension}"


@dataclass
class MediaMetadata:
    """The sub-object `mediaMetadata` of /tracks and /albums endpoint responses.
    It represents the quality levels available for the album's songs. These
    quality levels are determined by the client device type, the TIDAL account
    level, the country code (read: licensing), the device's quality settings,
    and, perhaps, the device's network connectivity conditions."""

    tags: List[str]


@dataclass(frozen=True)
class TrackAlbum:
    id: int
    title: str
    cover: str = field(repr=None)


@dataclass
class TracksEndpointResponseJSON(dataclass_wizard.JSONWizard):
    """Response from the TIDAL API, tracks/{TRACKID} endpoint.If the params and
    headers are correctly specified, the API returns metadata of the available
    version of the audio track, including audio quality, track title, ISRC,
    track artists, album, track number, duration, etc."""

    id: int = field(repr=False)
    title: str
    duration: int  # seconds
    replay_gain: float = field(repr=False)
    peak: float = field(repr=False)
    track_number: int
    volume_number: int
    version: Optional[str]
    copyright: str = field(repr=False)
    url: str
    isrc: str = field(repr=False)
    explicit: bool
    audio_quality: str = field(repr=False)
    audio_modes: List[str] = field(repr=False)
    media_metadata: "MediaMetadata"
    artist: "Artist"
    artists: List["Artist"]
    album: "TrackAlbum"

    def __post_init__(self):
        name: str = (
            self.title.replace("/", "_")
            .replace("|", "_")
            .replace(":", " -")
            .replace('"', "")
        )
        self.name: str = name if self.version is None else f"{name} ({self.version})"


@dataclass
class AlbumsEndpointResponseJSON(dataclass_wizard.JSONWizard):
    """This dataclass is the `dataclass-wizard`-generated class that represents
    the JSON response from https://api.tidal.com/v1/albums/<ALBUMID>"""

    id: int = field(repr=False)
    title: str
    duration: int
    number_of_tracks: int
    number_of_volumes: int = field(repr=False)
    release_date: date
    copyright: str = field(repr=False)
    type: str
    version: Optional[str]
    url: str
    cover: str = field(repr=False)
    explicit: bool
    upc: Union[int, str]
    audio_quality: str
    audio_modes: List[str]
    media_metadata: "MediaMetadata" = field(repr=False)
    artist: "Artist" = field(repr=False)
    artists: List["Artist"]

    def __post_init__(self):
        self.cover_url: str = IMAGE_URL % f"{self.cover.replace('-', '/')}/1280x1280"
        self.name: str = (
            self.title.replace("/", "_")
            .replace("|", "_")
            .replace(":", " -")
            .replace('"', "")
        )


@dataclass(frozen=True)
class SubscriptionEndpointResponseJSONSubscription:
    type: str
    offline_grace_period: int


@dataclass(frozen=True)
class SubscriptionEndpointResponseJSON(dataclass_wizard.JSONWizard):
    start_date: datetime
    valid_until: datetime
    status: str
    subscription: "SubscriptionEndpointResponseJSONSubscription"
    highest_sound_quality: str
    premium_access: bool
    can_get_trial: bool
    payment_type: str
    payment_overdue: bool


@dataclass(frozen=True)
class AlbumsItemsResponseJSONItem:
    """A sub-object of the response from the TIDAL API endpoint
    /albums/<ID>/items. It simply denotes the type of item, which is surely
    going to be 'track', and the same object that is returned from the TIDAL
    API /tracks endpoint."""

    item: "TracksEndpointResponseJSON"
    type: str  # "track"


@dataclass(frozen=True)
class AlbumsItemsResponseJSON(dataclass_wizard.JSONWizard):
    """This class represents the JSON response from the TIDAL API endpoint
    /albums/<ID>/items. It is a list of TracksEndpointResponseJSON objects,
    with a bit of metadata based on the query parameters (offset and limit;
    i.e. pagination logic)."""

    limit: int = field(repr=None)
    offset: int = field(repr=None)
    total_number_of_items: int
    items: List["AlbumsItemsResponseJSONItem"]


@dataclass(frozen=True)
class AlbumsReviewResponseJSON(dataclass_wizard.JSONWizard):
    """This class represents the JSON response from the TIDAL API endpoint
    /albums/<ID>/review."""

    source: str
    last_updated: Annotated[
        datetime, dataclass_wizard.Pattern("%Y-%m-%dT%H:%M:%S.%f%z")
    ]
    text: str = field(repr=None)
    summary: str = field(repr=None)


@dataclass(frozen=True)
class Contributor:
    """The response from the TIDAL API endpoint /tracks/<ID>/credits is
    an array of objects, one of the attributes of which is modeled by
    this class. It is simply the name of a contributor to a track,
    and possibly the numerical TIDAL resource ID of that contributor."""

    name: str
    id: Optional[int] = field(repr=False, default=None)


@dataclass(frozen=True)
class Credit:
    """The response from the TIDAL API endpoint /tracks/<ID>/credits is
    an array of objects modeled by this class. It has an attribute,
    `type`, which is one of the roles a person or entity has in the
    creation of a song/album: Composer, Lyricist, Producer, Mixer,
    Engineer, etc. The `contributors` attribute is an array of Name
    and, optionally, TIDAL resource ID for the role"""

    type: str
    contributors: List["Contributor"] = field(repr=False)


@dataclass
class TracksCreditsResponseJSON(dataclass_wizard.JSONWizard):
    """The response from the TIDAL API endpoint /tracks/<ID>/credits
    is modeled by this class."""

    credits: List["Credit"]

    def get_credit(self, type_: str) -> Optional["Credit"]:
        """Given a contributor type (e.g. Lyricist, Composer),
        go through the `credits` attribute, returning the `Credit` object
        for the given contributor type if it exists"""
        _credit = None
        try:
            _credit = next(c for c in self.credits if c.type == type_)
        except StopIteration:
            logger.debug(f"There are no credits of type {type_} for this track")
        finally:
            return _credit

    def get_contributors(self, type_: str) -> Optional[Tuple[str]]:
        """Given a contributor type (e.g. Lyricist, Composer),
        go through the `credits` attribute: for each Credit
        object in `self.credits`, if there is a Credit with
        `type` attribute matching `type_` argument, then return
        the `name` attribute for each Contributor object in
        `Credit.contributors`"""
        _credit: Optional["Credit"] = self.get_credit(type_)
        if _credit is not None:
            return tuple(c.name for c in _credit.contributors)
        else:
            return

    def __post_init__(self):
        """Try to parse the various Contributors to top-level
        attributes of this class"""
        self.composer: Optional[Tuple[str]] = self.get_contributors("Composer")
        self.engineer: Optional[Tuple[str]] = self.get_contributors("Engineer")
        self.lyricist: Optional[Tuple[str]] = self.get_contributors("Lyricist")
        self.mixer: Optional[Tuple[str]] = self.get_contributors("Mix Engineer")
        self.producer: Optional[Tuple[str]] = self.get_contributors("Producer")
        self.remixer: Optional[Tuple[str]] = self.get_contributors("Remixer")
        self.piano: Optional[Tuple[str]] = self.get_contributors("Piano")


@dataclass(frozen=True)
class TracksLyricsResponseJSON(dataclass_wizard.JSONWizard):
    """The response from the TIDAL API endpoint /tracks/<ID>/lyrics
    is modeled by this class."""

    track_id: int
    lyrics_provider: str
    provider_commontrack_id: str
    provider_lyrics_id: str
    lyrics: str
    subtitles: str
    is_right_to_left: bool


@dataclass
class ArtistsEndpointResponseJSON(dataclass_wizard.JSONWizard):
    """The response from the TIDAL API endpoint /artists is
    modeled by this class"""

    id: int
    name: str
    artist_types: List[str]
    url: str
    picture: Optional[str] = field(repr=False, default=None)

    def picture_url(self, dimension: int = 750) -> Optional[str]:
        if self.picture is None:
            return
        elif len(self.picture) != 36 or self.picture.count("-") != 4:
            # Should be a UUID
            return
        else:
            _picture = self.picture.replace("-", "/")
            return IMAGE_URL % f"{_picture}/{dimension}x{dimension}"


@dataclass(frozen=True)
class ArtistsAlbumsResponseJSON(dataclass_wizard.JSONWizard):
    """The response from the TIDAL API endpoint /artists/<ID>/albums
    is modeled by this class"""

    limit: int
    offset: int
    total_number_of_items: int
    items: List["AlbumsEndpointResponseJSON"]


@dataclass(frozen=True)
class ArtistsVideosResponseJSON(dataclass_wizard.JSONWizard):
    """The response from the TIDAL API endpoint /artists/<ID>/videos
    is modeled by this class"""

    limit: int
    offset: int
    total_number_of_items: int
    items: List["VideosEndpointResponseJSON"]


@dataclass(frozen=True)
class ArtistsBioResponseJSON(dataclass_wizard.JSONWizard):
    """The response from the TIDAL API endpoint /artists/<ID>/bio
    is modeled by this class."""

    source: str
    last_updated: Annotated[
        datetime, dataclass_wizard.Pattern("%Y-%m-%dT%H:%M:%S.%f%z")
    ]
    text: str = field(repr=None)
    summary: str = field(repr=None)


@dataclass
class VideosEndpointStreamResponseJSON(dataclass_wizard.JSONWizard):
    """Response from the TIDAL API's videos/<VIDEO_ID> stream
    endpoint. The params and headers, if correctly specified, return the
    manifest of the video to be streamed. The manifest is a base64-encoded
    JSON object containing a .m3u8 URL"""

    video_id: int
    stream_type: str  # ON_DEMAND
    video_quality: VideoQualityType
    manifest: str = field(repr=False)
    manifest_mime_type: str = field(repr=False)

    def __post_init__(self):
        self.manifest_bytes: bytes = base64.b64decode(self.manifest)


@dataclass
class VideosEndpointResponseJSON(dataclass_wizard.JSONWizard):
    """Response from the TIDAL API, videos/<VIDEOID> endpoint.If the params and
    headers are correctly specified, the API returns metadata of the available
    version of the (music) video, including video quality, video title, date,
    video artists, duration, etc."""

    id: int = field(repr=False)
    title: str
    volume_number: int
    track_number: int
    release_date: Annotated[
        datetime, dataclass_wizard.Pattern("%Y-%m-%dT%H:%M:%S.%f%z")
    ]
    duration: int  # seconds
    quality: str
    explicit: bool
    type: str
    artist: "Artist"
    artists: List["Artist"]

    def __post_init__(self):
        self.name: str = (
            self.title.replace("/", "_")
            .replace("|", "_")
            .replace(":", " -")
            .replace('"', "")
        )


@dataclass(frozen=True)
class VideoContributor:
    """The response from the TIDAL API endpoint /videos/<ID>/credits has
    an array of objects, each of which is modeled by this class.
    It is simply the name of a contributor to a video, and the role of that
    contributor."""

    name: str
    role: str


@dataclass
class VideosContributorsResponseJSON(dataclass_wizard.JSONWizard):
    """The response from the TIDAL API endpoint /videos/<ID>/contributors
    is modeled by this class."""

    limit: int
    offset: int
    total_number_of_items: int
    items: List["VideoContributor"]

    def get_role(self, role: str) -> Optional[Tuple["VideoContributor"]]:
        """Given a contributor role (e.g. Composer, Film Director), go through
        `self.items` object, returning the `VideoContributor` object(s)
        for the given contributor type if there are any"""
        role_contributors = tuple(vc for vc in self.items if vc.role == role)
        try:
            role_contributors[0]
        except IndexError:
            logger.debug(f"There are no credits of type '{role}' for this video")
            return
        else:
            return role_contributors

    def get_contributors(self, role: str) -> Optional[Tuple[str]]:
        """Given a contributor role (e.g. Lyricist, Composer),
        return a tuple of all the names of the contributors
        """
        vcs: Optional[Tuple["VideoContributor"]] = self.get_role(role)
        if vcs is not None:
            return tuple(vc.name for vc in vcs)
        else:
            return

    def __post_init__(self):
        """Try to parse the various Contributors to top-level
        attributes of this class"""
        self.composer: Optional[Tuple[str]] = self.get_contributors("Composer")
        self.director: Optional[Tuple[str]] = self.get_contributors("Director")
        self.film_director: Optional[Tuple[str]] = self.get_contributors(
            "Film Director"
        )
        self.film_producer: Optional[Tuple[str]] = self.get_contributors(
            "Film Producer"
        )
        self.lyricist: Optional[Tuple[str]] = self.get_contributors("Lyricist")
        self.mastering_engineer: Optional[Tuple[str]] = self.get_contributors(
            "Mastering Engineer"
        )
        self.producer: Optional[Tuple[str]] = self.get_contributors("Producer")
        self.video_producer: Optional[Tuple[str]] = self.get_contributors(
            "Video Producer"
        )


@dataclass
class PlaylistsEndpointResponseJSON(dataclass_wizard.JSONWizard):
    """Response from the TIDAL API, videos/<VIDEOID> endpoint.If the params and
    headers are correctly specified, the API returns metadata of the available
    version of the (music) video, including video quality, video title, date,
    video artists, duration, etc."""

    uuid: str = field(repr=False)
    title: str
    number_of_tracks: int
    number_of_videos: int
    description: str
    created: Annotated[datetime, dataclass_wizard.Pattern("%Y-%m-%dT%H:%M:%S.%f%z")]
    type: str
    public_playlist: bool
    url: str
    square_image: str  # UUID v4


class TidalResource:
    """Parent class to subclasses representing different TIDAL music
    service objects; e.g. Track, Album. This class is not meant to be
    instantiated itself: rather, its purpose is to pre-populate its
    subclasses with the `match_url` method."""

    def __init__(self, pattern: Optional[str] = None, url: Optional[str] = None):
        self.pattern = pattern
        self.url = url

    def match_url(self) -> Optional[Union[int, str]]:
        _match: re.Match = re.match(self.pattern, self.url, re.IGNORECASE)
        try:
            _id: str = _match.groups()[0]
        except AttributeError:
            return
        else:
            return _id


@dataclass
class TidalAlbum(TidalResource):
    """Class representing a TIDAL album. Its main purpose is the
    __post_init__ checking process"""

    url: str

    def __post_init__(self):
        self.pattern: str = (
            r"http(?:s)?://(?:listen\.)?tidal\.com/(?:browse/)?album/(\d{5,9})(?:.*?)?"
        )
        _id = self.match_url()

        if _id is None:
            raise ValueError(f"'{self.url}' is not a valid TIDAL album URL")
        else:
            self.tidal_id = int(_id)
            logger.info(f"TIDAL album ID parsed from input: {self.tidal_id}")


@dataclass
class TidalArtist(TidalResource):
    """Class representing a TIDAL artist. Its main purpose is the
    __post_init__ checking process"""

    url: str

    def __post_init__(self):
        self.pattern: str = (
            r"http(?:s)?://(?:listen\.)?tidal\.com/(?:browse/)?artist/(\d{7,9})(?:.*?)?"
        )
        _id = self.match_url()

        if _id is None:
            raise ValueError(f"'{self.url}' is not a valid TIDAL album URL")
        else:
            self.tidal_id = int(_id)
            logger.info(f"TIDAL album ID parsed from input: {self.tidal_id}")


@dataclass
class TidalMix(TidalResource):
    url: str

    def __post_init__(self):
        self.pattern: str = (
            r"http(?:s)?://(?:listen\.)?tidal\.com/(?:browse/)?mix/(\w{30})(?:.*?)?"
        )
        _id = self.match_url()

        if _id is None:
            raise ValueError(f"'{self.url}' is not a valid TIDAL mix URL")
        else:
            self.tidal_id = _id
            logger.info(f"TIDAL mix ID parsed from input: {self.tidal_id}")


@dataclass
class TidalTrack(TidalResource):
    """Class representing a TIDAL track. Its main purpose is the
    __post_init__ checking process"""

    url: str

    def __post_init__(self):
        self.pattern: str = r"http(?:s)?://(?:listen\.)?tidal\.com/(?:browse/)?(?:album/\d{7,9}/)?track/(\d{7,9})(?:.*?)?"
        _id = self.match_url()

        if _id is None:
            raise ValueError(f"'{self.url}' is not a valid TIDAL track URL")
        else:
            self.tidal_id = int(_id)
            logger.info(f"TIDAL track ID parsed from input: {self.tidal_id}")


@dataclass
class TidalPlaylist(TidalResource):
    """Class representing a TIDAL playlist. Its main purpose is the
    __post_init__ checking process"""

    url: str

    def __post_init__(self):
        self.pattern: str = (
            r"http(?:s)?://(?:listen\.)?tidal\.com/(?:browse/)?playlist/"
            r"([0-9a-f]{8}\-[0-9a-f]{4}\-4[0-9a-f]{3}\-[89ab][0-9a-f]{3}\-[0-9a-f]{12})(?:.*?)?"
        )

        _id = self.match_url()

        if _id is None:
            raise ValueError(f"'{self.url}' is not a valid TIDAL playlist URL")
        else:
            self.tidal_id = _id
            logger.info(f"TIDAL playlist ID parsed from input: {self.tidal_id}")


@dataclass
class TidalVideo(TidalResource):
    """Class representing a TIDAL video. Its main purpose is the
    __post_init__ checking process"""

    url: str

    def __post_init__(self):
        self.pattern: str = (
            r"http(?:s)?://(?:listen\.)?tidal\.com/(?:browse/)?video/(\d{7,9})(?:.*?)?"
        )
        _id = self.match_url()

        if _id is None:
            raise ValueError(f"'{self.url}' is not a valid TIDAL video URL")
        else:
            self.tidal_id = int(_id)
            logger.info(f"TIDAL video ID parsed from input: {self.tidal_id}")


def match_tidal_url(input_str: str) -> Optional[TidalResource]:
    """Attempt to match the `input_str` to either the URL of a track or an
    album in the Tidal API service. Returns None if `input_str` matches
    neither, otherwise a subclass of TidalResource corresponding to the
    parsed input_str type
    """
    resource_match: Optional[TidalResource] = None
    tidal_resources: Tuple[TidalResource] = (
        TidalTrack, TidalAlbum, TidalVideo, TidalPlaylist, TidalMix, TidalArtist
    )
    for T in tidal_resources:
        try:
            resource_match = T(input_str)
        except ValueError as v:
            logger.debug(v)
            continue
        else:
            return resource_match



import json
import logging
from requests import Session
from typing import Dict, List, Optional, Union

from .models import VideosEndpointStreamResponseJSON

import m3u8

logger = logging.getLogger(__name__)


class TidalM3U8Exception(Exception):
    pass


# https://github.com/globocom/m3u8#using-different-http-clients
class RequestsClient:
    """A custom class to pass to the m3u8.load() function"""

    def __init__(self, session: Session):
        self.session = session

    def download(
        self, url: str, timeout: Optional[int] = None, headers={}, verify_ssl=True
    ):
        p: Dict[str, None] = {k: None for k in self.session.params}
        with self.session.get(url=url, timeout=timeout, params=p) as response:
            return response.text, response.url


def playlister(
    session: Session, vesrj: Optional[VideosEndpointStreamResponseJSON]
) -> m3u8.M3U8:
    """Attempts to parse a VideosEndpointStreamResponseJSON object into an
    m3u8.M3U8 object. Requires fetching HTTP(s) resources, so takes a
    requests.Session object as an argument. If error occurs, raises
    TidalM3U8Exception"""
    em_three_you_ate: Optional[m3u8.M3U8] = None
    if vesrj.manifest_mime_type == "application/vnd.tidal.emu":
        try:
            manifest: Dict[str, str] = json.loads(vesrj.manifest_bytes)
        except json.decoder.JSONDecodeError:
            raise TidalM3U8Exception(
                f"Expected an HLS spec. in JSON format for video {vesrj.video_id}"
            )
        else:
            mt: Optional[str] = manifest.get("mimeType")
            url: Optional[str] = manifest.get("urls", [None])[0]

        if (
            (mt is None)
            or (mt != "application/vnd.apple.mpegurl")
            or (url is None)
            or (".m3u8" not in url)
        ):
            raise TidalM3U8Exception(
                f"Manifest for video {vesrj.video_id}, video mode "
                f"{vesrj.video_quality} does not make available an "
                "M3U8 file"
            )

        em_three_you_ate: m3u8.M3U8 = m3u8.load(
            url, http_client=RequestsClient(session=session)
        )
    return em_three_you_ate


def variant_streams(
    m3u8: m3u8.M3U8, return_urls: bool = False
) -> Optional[Union[m3u8.Playlist, List[str]]]:
    """By default, return the highest-bandwidth option of m3u8.playlists
    as an m3u8.Playlist object. If return_urls, then returns the object's
    .files attribute, which is a list of strings. N.b. if m3u8.is_variant
    is False, then return None as there are no variant streams."""
    if not m3u8.is_variant:
        return
    playlist: m3u8.Playlist = max(m3u8.playlists, key=lambda p: p.stream_info.bandwidth)
    if return_urls:
        _m3u8: m3u8.M3U8 = m3u8.load(playlist)
        return _m3u8.files
    else:
        return playlist



from dataclasses import dataclass
import json
import logging
from pathlib import Path
import shutil
import sys
from types import SimpleNamespace
from typing import Dict, List, Optional, Set, Tuple, Union

from requests import HTTPError, Session

from .media import AudioFormat
from .models import (
    PlaylistsEndpointResponseJSON,
    TracksEndpointResponseJSON,
    VideosEndpointResponseJSON,
)
from .track import Track
from .utils import TIDAL_API_URL
from .video import Video

logger = logging.getLogger("__name__")

# union type for type hinting
MixItem = Optional[Union["TracksEndpointResponseJSON", "VideosEndpointResponseJSON"]]


@dataclass
class Mix:
    mix_id: str

    def __post_init__(self):
        self.mix_dir: Optional[Path] = None
        self.mix_cover_saved: bool = False

    def get_metadata(self, session: Session):
        """Request from TIDAL API /playlists endpoint"""
        self.metadata: Optional[PlaylistsEndpointResponseJSON] = request_mixes(
            session=session, mix_id=self.mix_id
        )
        self.name = (
            self.metadata.title.replace("/", "_")
            .replace("|", "_")
            .replace(":", " -")
            .replace('"', "")
            .replace("..", "")
        )

    def set_items(self, session: Session):
        """Uses data from TIDAL API /mixes/items endpoint to
        populate self.items"""
        mix_items: Optional[MixesItemsResponseJSON] = get_mix(
            session=session, mix_id=self.mix_id
        )
        if mix_items is None:
            self.items = tuple()
        else:
            self.items: Tuple[Optional[MixItem]] = tuple(mix_items.items)

    def set_dir(self, out_dir: Path):
        """Populates self.mix_dir based on self.name, self.mix_id"""
        mix_substring: str = f"{self.name} [{self.mix_id}]"
        self.mix_dir: Path = out_dir / "Mixes" / mix_substring
        self.mix_dir.mkdir(parents=True, exist_ok=True)

    def save_cover_image(self, session: Session, out_dir: Path):
        """Requests self.metadata.image and attempts to write it to disk"""
        if self.mix_dir is None:
            self.set_dir(out_dir=out_dir)
        self.cover_path: Path = self.mix_dir / "cover.jpg"
        if not self.cover_path.exists():
            with session.get(
                url=self.metadata.image, params={k: None for k in session.params}
            ) as r:
                (self.mix_dir / "cover.jpg").write_bytes(r.content)

            self.mix_cover_saved = True
        else:
            self.mix_cover_saved = True

    def get_items(self, session: Session, audio_format: AudioFormat):
        """Using either Track.get() or Video.get(), attempt to request
        the data for each track or video in self.items"""
        if len(self.items) == 0:
            return
        tracks_videos: list = [None] * len(self.items)
        for i, item in enumerate(self.items):
            if item is None:
                tracks_videos[i] = None
                continue
            elif isinstance(item, TracksEndpointResponseJSON):
                track: Track = Track(track_id=item.id)
                track.get(
                    session=session,
                    audio_format=audio_format,
                    out_dir=self.mix_dir,
                    metadata=item,
                )
                tracks_videos[i] = track
            elif isinstance(item, VideosEndpointResponseJSON):
                video: Video = Video(video_id=item.id)
                video.get(
                    session=session,
                    out_dir=self.mix_dir,
                    metadata=item,
                )
                tracks_videos[i] = video
            else:
                tracks_videos[i] = None
                continue
        else:
            self.tracks_videos: Tuple[
                Tuple[int, Optional[Union[Track, Video]]]
            ] = tuple(tracks_videos)
        return tracks_videos

    def flatten_mix_dir(self):
        """When self.get_items() is called, the tracks and/or videos in
        self.items are downloaded using their self-contained .get() logic;
        this means that they will be downloaded to albums. This function
        "flattens" self.mix_dir, meaning that it moves all downloaded
        audio and video files to self.mix_dir, and removes the various
        subdirectories created"""
        files: List[Dict[int, Optional[str]]] = [None] * len(self.tracks_videos)
        if len(self.tracks_videos) == 0:
            return
        subdirs: Set[Path] = set()

        for i, tv in enumerate(self.tracks_videos, 1):
            if getattr(tv, "outfile") is None:
                try:
                    getattr(tv, "album_dir")
                except AttributeError:
                    pass
                else:
                    subdirs.add(tv.album_dir)
                    subdirs.add(tv.album_dir.parent)
                files[i - 1] = {i: None}
                continue

            _path: Optional[Path] = Path(tv.outfile) if tv is not None else None
            # if the item never got turned into a track or video
            if _path is None:
                files[i - 1] = {i: None}
                continue

            # if the track or video didn't download
            if _path.exists():
                if _path.stat().st_size == 0:
                    files[i - 1] = {i: None}
                    continue
            else:
                files[i - 1] = {i: None}
                continue

            # otherwise, move files and clean up
            if isinstance(tv, Track):
                new_path: Path = self.mix_dir / f"{i:03d} - {tv.trackname}"
                new_path.write_bytes(_path.read_bytes())
                _path.unlink()
                files[i - 1] = {i: str(new_path.absolute())}
            elif isinstance(tv, Video):
                new_path: Path = self.mix_dir / f"{i:03d} - {_path.name}"
                new_path.write_bytes(_path.read_bytes())
                _path.unlink()
                files[i - 1] = {i: str(new_path.absolute())}
        else:
            self.files: List[Dict[int, Optional[str]]] = files

        # Find all subdirectories written to
        subdirs: Set[Path] = set()
        for tv in self.tracks_videos:
            if isinstance(tv, Track):
                try:
                    getattr(tv, "album_dir")
                except AttributeError:
                    pass
                else:
                    subdirs.add(tv.album_dir)
                    subdirs.add(tv.album_dir.parent)
            elif isinstance(tv, Video):
                subdirs.add(tv.artist_dir)

        # Copy all artist images, artist bio JSON files out
        # of subdirs
        artist_images: Set[Path] = set()
        for subdir in subdirs:
            for p in subdir.glob("*.jpg"):
                if p.name == "cover.jpg":
                    continue
                artist_images.add(p)
        else:
            for artist_image_path in artist_images:
                if artist_image_path.exists():
                    shutil.copyfile(
                        artist_image_path.absolute(),
                        self.mix_dir / artist_image_path.name,
                    )

        artist_bios: Set[Path] = set()
        for subdir in subdirs:
            for p in subdir.glob("*bio.json"):
                artist_bios.add(p)
        else:
            for artist_bio_path in artist_bios:
                if artist_bio_path.exists():
                    shutil.copyfile(
                        artist_bio_path.absolute(),
                        self.mix_dir / artist_bio_path.name,
                    )

        # Remove all subdirs
        for subdir in subdirs:
            if subdir.exists():
                shutil.rmtree(subdir)
        else:
            return self.mix_dir

    def dumps(self):
        return json.dumps(self.files)

    def dump(self, fp=sys.stdout):
        json.dump(self.files, fp)

    def get(self, session: Session, audio_format: AudioFormat, out_dir: Path):
        """The main method of this class, executing a number of other methods
        in a row:
          - self.get_metadata()
          - self.set_items()
          - self.set_dir()
          - self.save_cover_image()
          - self.get_items()
          - self.flatten_playlist_dir()
        """
        self.get_metadata(session)
        self.set_items(session)
        self.set_dir(out_dir)
        self.save_cover_image(session, out_dir)
        try:
            self.save_description()
        except Exception:
            pass

        _get_items = self.get_items(session, audio_format)
        if _get_items is None:
            logger.critical(f"Could not retrieve mix with ID '{self.mix_id}'")
            return
        self.flatten_mix_dir()
        logger.info(f"Mix files written to '{self.mix_dir}'")


class TidalMixException(Exception):
    pass


def request_mixes(session: Session, mix_id: str) -> Optional[SimpleNamespace]:
    """Request from TIDAL API /pages/mix endpoint."""
    url: str = f"{TIDAL_API_URL}/pages/mix?mixId={mix_id}"
    kwargs: dict = {"url": url}
    kwargs["params"] = {"deviceType": "PHONE"}
    kwargs["headers"] = {"Accept": "application/json"}

    logger.info(f"Requesting from TIDAL API: mixes/{mix_id}/items")
    with session.get(**kwargs) as resp:
        try:
            resp.raise_for_status()
        except HTTPError as he:
            if resp.status_code == 404:
                logger.warning(
                    "404 Client Error: not found for TIDAL API endpoint pages/mix"
                )
            else:
                logger.exception(he)
            return

        d = dict()
        d["title"] = resp.json().get("title")
        d["description"] = resp.json().get("rows")[0]["modules"][0]["mix"]["subTitle"]
        d["image"] = (
            resp.json()
            .get("rows", [{}])[0]
            .get("modules")[0]["mix"]["images"]["LARGE"]["url"]
        )

        logger.debug(
            f"{resp.status_code} response from TIDAL API to request: pages/mix"
        )
        return SimpleNamespace(**d)


def request_mix_items(session: Session, mix_id: str) -> Optional[Dict]:
    """Request from TIDAL API /mixes/items endpoint."""
    url: str = f"{TIDAL_API_URL}/mixes/{mix_id}/items"
    kwargs: dict = {"url": url}
    kwargs["params"] = {"limit": 100}
    kwargs["headers"] = {"Accept": "application/json"}

    data: Optional[dict] = None
    logger.info(f"Requesting from TIDAL API: mixes/{mix_id}/items")
    with session.get(**kwargs) as resp:
        try:
            resp.raise_for_status()
        except HTTPError as he:
            if resp.status_code == 404:
                logger.warning(
                    f"404 Client Error: not found for TIDAL API endpoint mixes/{mix_id}/items"
                )
            else:
                logger.exception(he)
        else:
            data = resp.json()
            logger.debug(
                f"{resp.status_code} response from TIDAL API to request: mixes/{mix_id}/items"
            )
        finally:
            return data


@dataclass(frozen=True)
class MixesItemsResponseJSON:
    """The response from the TIDAL API endpoint /mixes/<ID>/items
    is modeled by this class."""

    limit: int
    offset: int
    total_number_of_items: int
    items: Tuple[
        Optional[Union["TracksEndpointResponseJSON", "VideosEndpointResponseJSON"]]
    ]


def mix_maker(
    mixes_response: Dict[str, Union[int, List[dict]]]
) -> "MixesItemsResponseJSON":
    """This function massages the response from the TIDAL API endpoint
    mixes/items into a format that MixesItemsResponseJSON.from_dict()
    can ingest. Returns a MixesItemsResponseJSON instance"""
    init_args: dict = {}
    init_args["limit"] = mixes_response.get("limit")
    init_args["offset"] = mixes_response.get("offset")
    init_args["total_number_of_items"] = mixes_response.get("totalNumberOfItems")

    items: Tuple[SimpleNamespace] = tuple(
        SimpleNamespace(**d) for d in mixes_response["items"]
    )
    if len(items) == 0:
        return

    mixes_items: List[
        Optional[Union["TracksEndpointResponseJSON", "VideosEndpointResponseJSON"]]
    ] = [None] * init_args["total_number_of_items"]

    for i, namespace in enumerate(items):
        if namespace.type == "track":
            try:
                mix_item = TracksEndpointResponseJSON.from_dict(namespace.item)
            except Exception as e:
                logger.warning(
                    f"TidalPlaylistException: unable to parse playlist item [i] "
                    f"with type '{namespace.type}'"
                )
                logger.debug(e)
                # value stays None
            else:
                mixes_items[i] = mix_item
        elif namespace.type == "video":
            try:
                mix_item = VideosEndpointResponseJSON.from_dict(namespace.item)
            except Exception as e:
                logger.warning(
                    f"TidalMixException: unable to parse mix item [i] "
                    f"with type '{namespace.type}'"
                )
                logger.debug(e)
                # value stays None
            else:
                mixes_items[i] = mix_item
        else:
            continue  # value stays None
    else:
        init_args["items"] = tuple(mixes_items)

    return MixesItemsResponseJSON(**init_args)


def get_mix(session: Session, mix_id: str) -> Optional["MixesItemsResponseJSON"]:
    """The pattern for mix items retrieval does not follow the
    requesting.request_* functions, hence its implementation here.
    """
    mixes_items_response_json: Optional["MixesItemsResponseJSON"] = None
    try:
        mixes_response: dict = request_mix_items(session=session, mix_id=mix_id)
        mixes_items_response_json: Optional["MixesItemsResponseJSON"] = mix_maker(
            mixes_response=mixes_response
        )
    except Exception as e:
        logger.exception(TidalMixException(e.args[0]))
    finally:
        return mixes_items_response_json



from dataclasses import dataclass
import json
import logging
from pathlib import Path
import sys
from typing import Dict, List, Optional

from .hls import playlister, variant_streams, TidalM3U8Exception
from .media import TAG_MAPPING, VideoFormat
from .models import (
    VideosContributorsResponseJSON,
    VideosEndpointResponseJSON,
    VideosEndpointStreamResponseJSON,
)
from .requesting import request_videos, request_video_contributors, request_video_stream
from .utils import temporary_file

import ffmpeg
import mutagen
import m3u8
from requests import Session

logger = logging.getLogger("__name__")


@dataclass
class Video:
    video_id: int

    def __post_init__(self):
        self.tags: dict = {}
        self.codec: str = "mp4"

    def get_metadata(self, session: Session):
        """Request from TIDAL API /videos endpoint"""
        self.metadata: Optional[VideosEndpointResponseJSON] = request_videos(
            session, self.video_id
        )

    def get_contributors(self, session: Session):
        """Request from TIDAL API /videos/contributors endpoint"""
        self.contributors: Optional[
            VideosContributorsResponseJSON
        ] = request_video_contributors(session, self.video_id)

    def get_stream(self, session: Session, video_format=VideoFormat.high):
        """Populates self.stream by requesting from TIDAL API
        /videos/playbackinfopostpaywall endpoint"""
        self.stream: Optional[VideosEndpointStreamResponseJSON] = request_video_stream(
            session, self.video_id, video_format.value
        )

    def get_m3u8(self, session: Session):
        """This method sets self.m3u8, an m3u8.M3U8 object
        following the HTTP Live Streaming specification; parsed from
        self.stream. I.e., self.get_stream() needs to have been executed
        before calling this method. N.b. self.m3u8 almost certainly will
        be a multivariant playlist, meaning further processing of its
        contents will be necessary."""
        self.m3u8: m3u8.Playlist = playlister(session=session, vesrj=self.stream)

    def set_urls(self):
        """This method uses self.m3u8, an m3u8.M3U8 object that is variant:
        (https://developer.apple.com/documentation/http-live-streaming/creating-a-multivariant-playlist)
        It retrieves the highest-quality .m3u8 in its .playlists attribute,
        and sets self.urls as the list of strings from that m3u8.Playlist"""
        # for now, just get the highest-bandwidth playlist
        playlist: m3u8.Playlist = variant_streams(self.m3u8)
        self.M3U8 = m3u8.load(playlist.uri)
        if self.M3U8 is None or len(self.M3U8.files) == 0:
            raise TidalM3U8Exception(
                f"HLS media segments are not available for video {self.video_id}"
            )
        self.urls: List[str] = self.M3U8.files

    def set_artist_dir(self, out_dir: Path):
        """Set self.artist_dir, which is the subdirectory of `out_dir`
        with name `self.metadata.artist.name`"""
        self.artist_dir: Path = out_dir / self.metadata.artist.name
        self.artist_dir.mkdir(parents=True, exist_ok=True)

    def set_filename(self, out_dir: Path):
        """Set self.filename, which is constructed from self.metadata.name
        and self.stream.video_quality"""
        self.filename: str = (
            f"{self.metadata.name} [{self.stream.video_quality}].{self.codec}"
        )

    def set_outfile(self):
        """Uses self.artist_dir and self.metadata and self.filename
        to craft the pathlib.Path object, self.outfile, that is a
        reference to where the track will be written on disk."""
        self.outfile: Path = self.artist_dir / self.filename

        if (self.outfile.exists()) and (self.outfile.stat().st_size > 0):
            logger.info(
                f"Video {str(self.outfile.absolute())} already exists "
                "and therefore will not be overwritten"
            )
            return
        else:
            return self.outfile

    def download(self, session: Session, out_dir: Path) -> Optional[Path]:
        """Requests the HLS video files that constitute self.video_id.
        Writes HLS bytes to a temporary file, then uses FFmpeg to write the
        video data to self.outfile"""
        if session.session_id is not None:
            download_headers: Dict[str, str] = {"sessionId": session.session_id}
        else:
            download_headers: dict = dict()
        download_params: Dict[str, None] = {k: None for k in session.params}
        # self.outfile should already have been set by self.set_outfile()
        logger.info(
            f"Writing video {self.video_id} to '{str(self.outfile.absolute())}'"
        )

        with temporary_file() as ntf:
            for u in self.urls:
                with session.get(
                    url=u, headers=download_headers, params=download_params
                ) as download_response:
                    if not download_response.ok:
                        logger.warning(f"Could not download {self}")
                    else:
                        ntf.write(download_response.content)
            else:
                ntf.seek(0)

            # will always be .mp4 because HLS
            ffmpeg.input(ntf.name, hide_banner=None, y=None).output(
                str(self.outfile.absolute()),
                vcodec="copy",
                acodec="copy",
                loglevel="quiet",
            ).run()

        logger.info(
            f"Video {self.video_id} written to '{str(self.outfile.absolute())}'"
        )
        return self.outfile

    def craft_tags(self):
        """Using the TAG_MAPPING dictionary, write the correct values of
        various metadata tags to the file. Videos are .mp4"""
        tags = dict()
        tag_map = {k: v["m4a"] for k, v in TAG_MAPPING.items()}

        tags[tag_map["artist"]] = ";".join((a.name for a in self.metadata.artists))
        tags[tag_map["artists"]] = [a.name for a in self.metadata.artists]
        tags[tag_map["comment"]] = f"https://tidal.com/browse/video/{self.video_id}"
        tags[tag_map["date"]] = str(self.metadata.release_date.date())
        tags[tag_map["title"]] = self.metadata.title

        for tag in {"composer", "director", "lyricist", "producer"}:
            try:
                _credits_tag = ";".join(getattr(self.contributors, tag))
            except (TypeError, AttributeError):  # NoneType problems
                continue
            else:
                tags[tag_map[tag]] = _credits_tag

        # Have to convert to bytes the values of the tags starting with '----'
        for k, v in tags.copy().items():
            if k.startswith("----"):
                if isinstance(v, str):
                    tags[k]: bytes = v.encode("UTF-8")
                elif isinstance(v, list):
                    tags[k]: List[bytes] = [s.encode("UTF-8") for s in v]

        self.tags: dict = {k: v for k, v in tags.items() if v is not None}

    def set_tags(self):
        """Instantiate a mutagen.File instance, add self.tags to it, and
        save it to disk"""
        self.mutagen = mutagen.File(self.outfile)
        self.mutagen.clear()
        self.mutagen.update(**self.tags)
        self.mutagen.save()

    def get(
        self,
        session: Session,
        out_dir: Path,
        metadata: Optional["VideosEndpointResponseJSON"] = None,
    ) -> Optional[str]:
        """The main method of this class. Executes a number of other methods
        in a row:
          - self.get_metadata()
          - self.get_contributors()
          - self.get_stream()
          - self.get_m3u8()
          - self.set_urls()
          - self.set_artist_dir()
          - self.set_filename()
          - self.set_outfile()
          - self.download()
          - self.craft_tags()
          - self.set_tags()
        """
        if metadata is None:
            self.get_metadata(session)
        else:
            self.metadata = metadata

        # check for 404 error with metadata
        if self.metadata is None:
            return None

        self.get_contributors(session)
        self.get_stream(session)
        if self.stream is None:
            return None
        self.get_m3u8(session)
        self.set_urls()
        self.set_artist_dir(out_dir)
        self.set_filename(out_dir)
        outfile: Optional[Path] = self.set_outfile()
        if outfile is None:
            return None

        if self.download(session, out_dir) is None:
            return None

        self.craft_tags()
        self.set_tags()
        return str(self.outfile.absolute())

    def dump(self, fp=sys.stdout):
        json.dump({self.metadata.title: str(self.outfile.absolute())}, fp)

    def dumps(self) -> str:
        return json.dumps({self.metadata.title: str(self.outfile.absolute())})
